# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

import unittest

import util


class TestSehEmulation(unittest.TestCase):

    def setUp(self):
        self.config = util.get_config()

        self.dispatch_script = ['Hello emulator\n',
                                'First access violation\r\n',
                                'First nested access violation\r\n',
                                'Second nested access violation\r\n',
                                'After access violations\r\n',
                                'In finally\r\n',
                                'Returning...\n']

    def test_seh_dispatch(self):
        self.config['exceptions']['dispatch_handlers'] = True
        data = util.get_test_bin_data('seh_test_x86.exe.xz')
        report = util.run_test(self.config, data)

        ep = report['entry_points']
        printfs = []
        for api in ep[0]['apis']:
            if '__stdio_common_vfprintf' in api['api_name']:
                printfs.append(api)

        fmt_strings = [p['args'][2] for p in printfs]
        self.assertEqual(len(fmt_strings), len(self.dispatch_script))
        for i, s in enumerate(fmt_strings):
            self.assertEqual(s, self.dispatch_script[i])

    def test_seh_without_dispatch(self):
        self.config['exceptions']['dispatch_handlers'] = False

        data = util.get_test_bin_data('seh_test_x86.exe.xz')
        report = util.run_test(self.config, data)

        ep = report['entry_points']
        printfs = []
        for api in ep[0]['apis']:
            if '__stdio_common_vfprintf' in api['api_name']:
                printfs.append(api)
                break

        self.assertEqual(1, len(printfs))
        error = ep[0]['error']

        self.assertEqual(error['type'], 'invalid_write')
        self.assertEqual(error['address'], '0x0')
        self.assertEqual(error['instr'], 'mov dword ptr [0], 0x14')


if __name__ == '__main__':
    unittest.main()



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

class KernelModule(object):

    def __init__(self):
        self.name = ''

    def get_mod_name(self):
        return self.name.lower()

    def ioctl(self, arch, code, inbuf):
        raise NotImplementedError



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

from speakeasy.struct import EmuStruct, Ptr
import ctypes as ct


class GUID(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.Data1 = ct.c_uint32
        self.Data2 = ct.c_uint16
        self.Data3 = ct.c_uint16
        self.Data4 = ct.c_uint8 * 8


class FWPM_DISPLAY_DATA0(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.name = Ptr
        self.description = Ptr


class FWP_VALUE0(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.type = ct.c_uint32
        self.data = Ptr


class FWP_BYTE_BLOB(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.size = ct.c_uint32
        self.data = Ptr


class FWPM_SUBLAYER0(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.subLayerKey = GUID
        self.displayData = FWPM_DISPLAY_DATA0
        self.flags = ct.c_uint32
        self.providerKey = GUID
        self.providerData = FWP_BYTE_BLOB
        self.weight = ct.c_uint16


class FWPS_CALLOUT1(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.calloutKey = GUID
        self.flags = ct.c_uint32
        self.classifyFn = Ptr
        self.notifyFn = Ptr
        self.flowDeleteFn = Ptr


class FWPM_CALLOUT0(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.calloutKey = GUID
        self.displayData = FWPM_DISPLAY_DATA0
        self.flags = ct.c_uint32
        self.providerKey = GUID
        self.providerData = FWP_BYTE_BLOB
        self.applicableLayer = GUID
        self.calloutId = ct.c_uint32


class FWPM_FILTER_CONDITION0(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.fieldKey = GUID
        self.matchType = ct.c_uint32
        self.conditionValue = FWP_VALUE0


class FWPM_ACTION0(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.type = ct.c_uint32
        self.filterType = GUID


class FWPM_FILTER0(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.filterKey = GUID
        self.displayData = FWPM_DISPLAY_DATA0
        self.flags = ct.c_uint32
        self.providerKey = GUID
        self.providerData = FWP_BYTE_BLOB
        self.layerKey = GUID
        self.subLayerKey = GUID
        self.weight = FWP_VALUE0
        self.numFilterConditions = ct.c_uint32
        self.filterCondition = Ptr
        self.action = FWPM_ACTION0
        self.providerContextKey = GUID
        self.reserved = Ptr
        self.filterId = ct.c_uint64
        self.effectiveWeight = FWP_VALUE0



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

import os
import json
import time
import logging
import argparse
import multiprocessing as mp

import speakeasy
from speakeasy import Speakeasy

import speakeasy.winenv.arch as e_arch


def get_logger():
    """
    Get the default logger for speakeasy
    """
    logger = logging.getLogger('speakeasy')
    if not logger.handlers:
        sh = logging.StreamHandler()
        logger.addHandler(sh)
        logger.setLevel(logging.INFO)

    return logger


def emulate_binary(q, exit_event, fpath, cfg, argv, do_raw, arch='',
        drop_path='', dump_path='', raw_offset=0x0, emulate_children=False):
    """
    Setup the binary for emulation
    """

    logger = get_logger()

    try:
        report = None
        se = Speakeasy(config=cfg, logger=logger, argv=argv, exit_event=exit_event)
        if do_raw:
            arch = arch.lower()
            if arch == 'x86':
                arch = e_arch.ARCH_X86
            elif arch in ('x64', 'amd64'):
                arch = e_arch.ARCH_AMD64
            else:
                raise Exception('Unsupported architecture: %s' % arch)

            sc_addr = se.load_shellcode(fpath, arch)
            se.run_shellcode(sc_addr, offset=raw_offset or 0)
        else:
            module = se.load_module(fpath)
            se.run_module(module, all_entrypoints=True,
                    emulate_children=emulate_children)

    finally:

        report = se.get_json_report()
        q.put(report)

        # If a memory dump was requested, do it now
        if dump_path:
            data = se.create_memdump_archive()
            logger.info('* Saving memory dump archive to %s' % (dump_path))
            with open(dump_path, 'wb') as f:
                f.write(data)

        if drop_path:
            data = se.create_file_archive()
            if data:
                logger.info('* Saving dropped files archive to %s' % (drop_path))
                with open(drop_path, 'wb') as f:
                    f.write(data)
            else:
                logger.info('* No dropped files found')


class Main(object):
    """
    Main class for emulation of Windows shellcode, user mode, and kernel mode binaries
    """
    def __init__(self, parser):
        args = parser.parse_args()
        self.target = args.target
        self.output = args.output
        self.dump_path = args.dump_path
        self.drop_files_path = args.drop_files_path
        self.config_path = args.config
        self.emulate_children = args.emulate_children
        self.cfg = None
        self.do_raw = args.do_raw
        self.raw_offset = args.raw_offset
        self.do_memtrace = args.do_memtrace
        self.module_dir = args.module_dir
        self.arch = args.arch
        self.timeout = 0
        self.argv = args.params
        self.logger = get_logger()

        if not self.config_path:
            self.config_path = './speakeasy/configs/default.json'
        if not os.path.isfile(self.config_path):
            self.config_path = os.path.join(os.path.dirname(speakeasy.__file__),
                                            'configs', 'default.json')
            if not os.path.isfile(self.config_path):
                parser.print_help()
                self.logger.error('[-] No emulator config file supplied')
                return

        with open(self.config_path, 'r') as f:
            self.cfg = json.load(f)
            if args.timeout:
                self.timeout = args.timeout
                self.cfg.update({'timeout': self.timeout})
                self.cfg.update({'max_api_count': self.timeout * 500})
            else:
                self.timeout = self.cfg.get('timeout', 0)

            if self.do_memtrace:
                analysis = self.cfg.get('analysis', {})
                # Override and enable memory tracing
                if analysis:
                    analysis['memory_tracing'] = True
                else:
                    self.cfg.update({'analysis': {'memory_tracing': True}})

            if self.module_dir:
                modules = self.cfg.get('modules', {})
                if modules:
                    modules['module_directory_x86'] = self.module_dir
                    modules['module_directory_x64'] = self.module_dir
                else:
                    self.cfg.update({'modules': {'module_directory_x86': self.module_dir,
                                                 'module_directory_x64': self.module_dir}})

        if self.target and not os.path.isfile(self.target):
            parser.print_help()
            self.logger.error('[-] Target file not found: %s' % (self.target))
            return

        if not self.target:
            parser.print_help()
            self.logger.error('[-] No target file supplied')
            return

        q = mp.Queue()
        evt = mp.Event()

        if args.no_mp:
            # Emulate within the current process, losing some control with execution but
            # allows us to debug speakeasy.
            emulate_binary(q, evt, args.target,
                           self.cfg, self.argv, self.do_raw, self.arch,
                           self.drop_files_path, self.dump_path,
                           raw_offset=self.raw_offset,
                           emulate_children=self.emulate_children)
            report = q.get()
        else:
            # We are using a child process here so we can maintain absolute control over its
            # execution
            p = mp.Process(target=emulate_binary,
                           args=(q, evt, args.target, self.cfg,
                                 self.argv, self.do_raw, self.arch,
                                 self.drop_files_path, self.dump_path),
                           kwargs={
                               "raw_offset": self.raw_offset,
                               "emulate_children": self.emulate_children,
                           })
            p.start()

            report = None
            start_time = time.time()
            while True:
                if self.timeout and self.timeout < (time.time() - start_time):
                    evt.set()
                    self.logger.error('* Child process timeout reached after %d seconds' %
                                      (self.timeout))
                    report = q.get(5)
                try:
                    report = q.get(timeout=1)
                    break
                except mp.queues.Empty:
                    if not p.is_alive():
                        break
                except KeyboardInterrupt:
                    evt.set()
                    self.logger.error('\n* User exited')
                    report = q.get(5)
                    break

        self.logger.info('* Finished emulating')

        if report:
            if self.output:
                self.logger.info('* Saving emulation report to %s' % (self.output))
                with open(self.output, 'w') as f:
                    f.write(report)


def main():
    """ speakeasy command line entrypoint """

    parser = argparse.ArgumentParser(description='Emulate a Windows binary with speakeasy')
    parser.add_argument('-t', '--target', action='store', dest='target',
                        required=False, help='Path to input file to emulate')
    parser.add_argument('-o', '--output', action='store', dest='output',
                        required=False, help='Path to output file to save report')
    parser.add_argument('-p', '--params', action='store', default=[],
                        nargs='*', dest='params', required=False,
                        help='Commandline parameters to supply to emulated '
                             'process (e.g. main(argv))')
    parser.add_argument('-c', '--config', action='store', dest='config',
                        required=False, help='Path to emulator config file')
    parser.add_argument('-m', '--mem-tracing', action='store_true', dest='do_memtrace',
                        required=False, help='Enables memory tracing.\n'
                                             'This will log all memory access by the '
                                             'sample but will impact speed')
    parser.add_argument('-r', '--raw', action='store_true', dest='do_raw',
                        required=False, help='Attempt to emulate file as-is '
                                             'with no parsing (e.g. shellcode)')
    parser.add_argument('--raw_offset', type=lambda s: int(s, 0x10), default=0,
                        required=False, help='When in raw mode, offset (hex) to start emulating')
    parser.add_argument('-a', '--arch', action='store', dest='arch',
                        required=False,
                        help='Force architecture to use during emulation (for '
                             'multi-architecture files or shellcode). '
                             'Supported archs: [ x86 | amd64 ]')
    parser.add_argument('-d', '--dump', action='store', dest='dump_path',
                        required=False, help='Path to store compressed memory dump package')
    parser.add_argument('-q', '--timeout', action='store', dest='timeout', type=int,
                        required=False, help='Emulation timeout in seconds (default 60 sec)')
    parser.add_argument('-z', '--dropped-files', action='store', dest='drop_files_path',
                        required=False, help='Path to store files created during emulation')
    parser.add_argument('-l', '--module-dir', action='store', dest='module_dir',
                        required=False, help='Path to directory containing loadable PE modules.\n'
                                             'When modules are parsed or loaded by samples,\n'
                                             'PEs from this directory will be loaded into the\n'
                                             'emulated address space')
    parser.add_argument('-k', '--emulate-children', action='store_true', dest='emulate_children',
                        required=False, help='Emulate any processes created with\n'
                                             'the CreateProcess APIs after the\n'
                                             'input file finishes emulating')
    parser.add_argument('--no-mp', action='store_true', dest='no_mp',
                        required=False, help='Run emulation in the current process to assist\n'
                                             'instead of a child process. Useful when debugging'
                                             'speakeasy itself (using pdb.set_trace()).\n')

    Main(parser)

if __name__ == "__main__":
    main()



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

import ctypes

import speakeasy.winenv.defs.windows.com as comdefs
import speakeasy.winenv.api.usermode.com_api as com_api
from speakeasy.errors import Win32EmuError


class COM(object):
    """
    The Component Object Model (COM) manager for the emulator. This will manage COM interfaces.
    """
    def __init__(self, config):
        super(COM, self).__init__()
        self.interfaces = {}
        self.config = config

    def get_interface(self, emu, ptr_size, name):
        """
        Get COM interface
        """
        iface = comdefs.IFACE_TYPES.get(name)
        if not iface:
            raise Win32EmuError('Invalid COM interface: %s' % (name))

        ci = comdefs.ComInterface(iface, name, ptr_size)
        com_ptr = emu.mem_map(emu.sizeof(ci.iface), tag='emu.COM.%s' % (name))
        ci.address = com_ptr

        fields = ci.iface.__dict__['__fields__']
        field_offset = 0
        for field in fields:
            field_name, field_obj = field
            # Determine if the field is an inherited interface (e.g., IUnknown)
            if issubclass(field_obj, ctypes.Structure):
                if field_name not in comdefs.IFACE_TYPES:
                    raise Win32EmuError('COM interface %s inherits unsupported interface %s' %
                                        (name, field_name))

                # Iterate inherited interface fields
                for subfield in field_obj._fields_:
                    subfield_name, subfield_type = subfield
                    if issubclass(subfield_type, (ctypes.c_uint32, ctypes.c_ulong,
                                                  ctypes.c_ulonglong)):
                        # Inherited inferface field is a method; hook if supported
                        method_name = '%s_%s' % (field_name, subfield_name)
                        if hasattr(com_api.ComApi, method_name):
                            method = getattr(com_api.ComApi, method_name)
                            addr = emu.add_callback(com_api.ComApi.name, method.__apihook__[0])
                            emu.mem_write(com_ptr + field_offset, addr.to_bytes(emu.get_ptr_size(),
                                          'little'))

                    field_offset += ptr_size
            elif issubclass(field_obj, (ctypes.c_uint32, ctypes.c_ulong, ctypes.c_ulonglong)):
                # Field is a method; hook if supported
                method_name = '%s_%s' % (name, field_name)
                if hasattr(com_api.ComApi, method_name):
                    method = getattr(com_api.ComApi, method_name)
                    addr = emu.add_callback(com_api.ComApi.name, method.__apihook__[0])
                    emu.mem_write(com_ptr + field_offset, addr.to_bytes(emu.get_ptr_size(),
                                  'little'))

                field_offset += ptr_size
            else:
                raise Win32EmuError('Invalid field type encountered for %s.%s' %
                                    (name, field_name))

        return ci



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

import base64
from .. import api

ERROR_MORE_DATA = 234

CRYPT_STRING_BASE64 = 1


class Crypt32(api.ApiHandler):

    """
    Implements exported functions from user32.dll
    """

    name = 'crypt32'
    apihook = api.ApiHandler.apihook
    impdata = api.ApiHandler.impdata

    def __init__(self, emu):

        super(Crypt32, self).__init__(emu)

        self.funcs = {}
        self.data = {}

        super(Crypt32, self).__get_hook_attrs__(self)

    @apihook('CryptStringToBinary', argc=7)
    def CryptStringToBinary(self, emu, argv, ctx={}):
        '''
        BOOL CryptStringToBinaryA(
        LPCSTR pszString,
        DWORD  cchString,
        DWORD  dwFlags,
        BYTE   *pbBinary,
        DWORD  *pcbBinary,
        DWORD  *pdwSkip,
        DWORD  *pdwFlags
        );
        '''

        cw = self.get_char_width(ctx)

        pszString, cchString, dwFlags, pbBinary, pcbBinary, pdwSkip, pdwFlags = argv

        if cchString:
            s = self.mem_read(pszString, cchString * cw)
        else:
            s = self.read_mem_string(pszString, cw)

        if dwFlags != CRYPT_STRING_BASE64:
            # self.logger.info("%s: currently unsupported flags (%08X)" % (api_name, dwFlags))
            return 1

        if type(s) != str:
            s = s.decode('utf8')

        argv[0] = s

        try:
            decoded = base64.b64decode(s)
        except Exception:
            return 0

        cbBinary = int.from_bytes(self.mem_read(pcbBinary, 4), 'little')
        out_len = len(decoded)

        if pbBinary == 0:
            return out_len

        if out_len > cbBinary:
            emu.set_last_error(ERROR_MORE_DATA)
            return 0

        self.mem_write(pbBinary, decoded)
        self.mem_write(pcbBinary, out_len.to_bytes(4, 'little'))

        if pdwSkip:
            self.mem_write(pdwSkip, b"\x00" * 4)

        return 1



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

import io
import os
import ntpath
import traceback
import shlex

import speakeasy.winenv.arch as _arch
from speakeasy.binemu import BinaryEmulator
from speakeasy.profiler import MemAccess

import speakeasy.common as common
from speakeasy.profiler import Run
import speakeasy.windows.common as winemu
import speakeasy.windows.objman as objman
from speakeasy.windows.regman import RegistryManager
from speakeasy.windows.fileman import FileManager
from speakeasy.windows.cryptman import CryptoManager
from speakeasy.windows.netman import NetworkManager
from speakeasy.windows.hammer import ApiHammer
from speakeasy.windows.driveman import DriveManager

import speakeasy.winenv.defs.nt.ddk as ddk
import speakeasy.winenv.defs.windows.windows as windef

from speakeasy.struct import EmuStruct
from speakeasy.errors import WindowsEmuError

# When disassembling, a minimum instruction size needs to be supplied
# This number is arbitrary and just needs to be large enough to cover
# the size of the current disasm target
DISASM_SIZE = 0x20


class WindowsEmulator(BinaryEmulator):
    """
    Base class providing emulation of all Windows modules and shellcode.
    This class is meant to provide overlapping functionality for both
    user mode and kernel mode samples.
    """

    def __init__(self, config, logger=None, exit_event=None, debug=False):
        super(WindowsEmulator, self).__init__(config, logger=logger)

        self.debug = debug
        self.arch = 0
        self.modules = []
        self.pic_buffers = []
        self.curr_run = None
        self.restart_curr_run = False
        self.curr_mod = None
        self.runs = []
        self.input = None
        self.exit_event = exit_event
        self.page_size = 4096
        self.ptr_size = None
        self.user_modules = []
        self.max_runs = 100

        self.sys_modules = []
        self.symbols = {}
        self.ansi_strings = []
        self.unicode_strings = []
        self.tmp_maps = []
        self.impdata_queue = []
        self.run_queue = []
        self.suspended_runs = []
        self.cd = ''
        self.emu_hooks_set = True
        self.api = None
        self.curr_process = None
        self.om = None
        self.dyn_imps = []
        self.callbacks = []
        self.mem_trace_hooks = []
        self.kernel_mode = False
        self.virtual_mem_base = 0x50000

        self.mem_tracing_enabled = False
        self.tmp_code_hook = None
        self.veh_handlers = []

        self.run_complete = False
        self.emu_complete = False
        self.global_data = {}
        self.processes = []
        # Child processes created by calls to CreateProcess
        # by any module. This is separate from self.processes in order
        # to not mix up config processes with child processes
        self.child_processes = []
        self.curr_thread = None
        self.curr_exception_code = 0
        self.prev_pc = 0
        self.unhandled_exception_filter = 0

        self.fs_addr = 0
        self.gs_addr = 0

        self.return_hook = winemu.EMU_RETURN_ADDR
        self.exit_hook = winemu.EXIT_RETURN_ADDR
        self._parse_config(config)

        self.wintypes = windef
        # OS resource managers
        self.regman = RegistryManager(self.get_registry_config())
        self.fileman = FileManager(config, self)
        self.netman = NetworkManager(config=self.get_network_config())
        self.driveman = DriveManager(config=self.get_drive_config())
        self.cryptman = CryptoManager()
        self.hammer = ApiHammer(self)

    def _parse_config(self, config):
        """
        Parse the emulation config file
        """
        def _normalize_image(img):
            # Normalize the architecture
            if img['arch'].lower() in ('x86', 'i386'):
                img['arch'] = _arch.ARCH_X86
            elif img['arch'].lower() in ('x64', 'amd64'):
                img['arch'] = _arch.ARCH_AMD64
            else:
                raise WindowsEmuError('Unsupported image arch: %s'
                                      % (img['arch']))

        super(WindowsEmulator, self)._parse_config(config)
        for umod in self.config_user_modules:
            for img in umod.get('images', []):
                _normalize_image(img)

        for proc in self.config_processes:
            for img in proc.get('images', []):
                _normalize_image(img)

        self.cd = self.config.get('current_dir', '')

        self.dispatch_handlers = self.exceptions.get('dispatch_handlers', True)
        self.mem_tracing_enabled = self.config_analysis.get('memory_tracing', False)
        self.do_strings = self.config_analysis.get('strings', False)
        self.registry_config = self.config.get('registry', {})
        self.modules_always_exist = self.config_modules.get('modules_always_exist', False)
        self.functions_always_exist = self.config_modules.get('functions_always_exist', False)

    def get_registry_config(self):
        """
        Get the registry settings specified in the registry section of the config file
        """
        return self.registry_config

    def on_run_complete(self):
        """
        Clean up after a run completes (implemented in the child class) since
        this may mean different things depending. This function will pop the
        next run from the run queue and emulate it.
        """
        # Implemented by a subclass (e.g. kernel/user mode emulators)
        raise NotImplementedError()

    def enable_code_hook(self):
        if not self.tmp_code_hook and not self.mem_tracing_enabled:
            self.tmp_code_hook = self.add_code_hook(cb=self._hook_code)

        if self.tmp_code_hook:
            self.tmp_code_hook.enable()

    def disable_code_hook(self):
        if self.tmp_code_hook:
            self.tmp_code_hook.disable()

    def _module_access_hook(self, emu, addr, size, ctx):
        symbol = self.get_symbol_from_address(addr)
        if symbol:
            mod_name, fn = symbol.split('.')
            self.handle_import_func(mod_name, fn)
            return True

    def set_mem_tracing_hooks(self):
        if not self.mem_tracing_enabled:
            return

        if len(self.mem_trace_hooks) > 0:
            return

        self.mem_trace_hooks = (
            self.add_code_hook(cb=self._hook_code),
            self.add_mem_read_hook(cb=self._hook_mem_read),
            self.add_mem_write_hook(cb=self._hook_mem_write)
        )

    def cast(self, obj, bytez):
        """
        Create a formatted structure from bytes
        """
        if not isinstance(obj, EmuStruct):
            raise WindowsEmuError('Invalid object for cast')
        return obj.cast(bytez)

    def _unset_emu_hooks(self):
        """
        Create a formatted structure from bytes
        """
        if self.emu_hooks_set:
            self.emu_eng.mem_map(winemu.EMU_RETURN_ADDR,
                                 winemu.EMU_RESERVE_SIZE)
        self.emu_hooks_set = False

    def file_open(self, path, create=False):
        """
        Open an emulated from using the file manager
        """
        return self.fileman.file_open(path, create)

    def pipe_open(self, path, mode, num_instances, out_size, in_size):
        """
        Open an emulated named pipe
        """
        return self.fileman.pipe_open(path, mode, num_instances, out_size, in_size)

    def does_file_exist(self, path):
        """
        Test if a file handler for a specified emulated file exists
        """
        return self.fileman.does_file_exist(path)

    def file_create_mapping(self, hfile, name, size, prot):
        """
        Create a memory mapping for an emulated file
        """
        return self.fileman.file_create_mapping(hfile, name, size, prot)

    def file_get(self, handle):
        """
        Get a file object from a handle
        """
        return self.fileman.get_file_from_handle(handle)

    def file_delete(self, path):
        """
        Delete a file
        """
        return self.fileman.delete_file(path)

    def pipe_get(self, handle):
        """
        Get a pipe object from a handle
        """
        return self.fileman.get_pipe_from_handle(handle)

    def get_file_manager(self):
        """
        Get the file emulation manager
        """
        return self.fileman

    def get_network_manager(self):
        """
        Get the network emulation manager
        """
        return self.netman

    def get_crypt_manager(self):
        """
        Get the crypto manager
        """
        return self.cryptman

    def get_drive_manager(self):
        """
        Get the drive manager
        """
        return self.driveman

    def reg_open_key(self, path, create=False):
        """
        Open or create a registry key in the emulation space
        """
        return self.regman.open_key(path, create)

    def reg_get_subkeys(self, hkey):
        """
        Get subkeys for a given registry key
        """
        return self.regman.get_subkeys(hkey)

    def reg_get_key(self, handle=0, path=''):
        """
        Get registry key by path or handle
        """
        if path:
            return self.regman.get_key_from_path(path)
        return self.regman.get_key_from_handle(handle)

    def reg_create_key(self, path):
        """
        Create a registry key
        """
        return self.regman.create_key(path)

    def _set_emu_hooks(self):
        """
        Unmap reserved memory space so we can handle events (e.g. import APIs,
        entry point returns, etc.)
        """
        if not self.emu_hooks_set:
            self.mem_unmap(winemu.EMU_RETURN_ADDR, winemu.EMU_RESERVE_SIZE)
            self.emu_hooks_set = True

    def add_run(self, run):
        """
        Add a run to the emulation run queue
        """
        self.run_queue.append(run)

    def _exec_next_run(self):
        """
        Execute the next run from the emulation queue
        """
        try:
            run = self.run_queue.pop(0)
        except IndexError:
            self.on_emu_complete()
            return None

        self.run_complete = False
        self.reset_stack(self.stack_base)
        return self._exec_run(run)

    def call(self, addr, params=[]):
        """
        Start emulating at the specified address
        """
        self.reset_stack(self.stack_base)
        run = Run()
        run.type = 'call_0x%x' % (addr)
        run.start_addr = addr
        run.args = params

        if not self.run_queue:
            self.add_run(run)
            self.start()
        else:
            self.add_run(run)

    def _exec_run(self, run):
        """
        Begin emulating the specified run
        """
        self.log_info("* exec: %s" % run.type)

        self.curr_run = run
        if self.profiler:
            self.profiler.add_run(run)

        self.runs.append(self.curr_run)

        stk_ptr = self.get_stack_ptr()

        self.set_func_args(stk_ptr, self.return_hook, *run.args)
        stk_ptr = self.get_stack_ptr()
        stk_map = self.get_address_map(stk_ptr)

        self.curr_run.stack = MemAccess(base=stk_map.base, size=stk_map.size)

        # Set the process context if possible
        if run.process_context:
            # Init a new peb if the process context changed:
            if run.process_context != self.get_current_process():
                self.alloc_peb(run.process_context)
            self.set_current_process(run.process_context)
        if run.thread:
            self.set_current_thread(run.thread)

        if not self.kernel_mode:
            # Reset the TIB data
            thread = self.get_current_thread()
            if thread:
                self.init_teb(thread, self.curr_process.get_peb())
                self.init_tls(thread)

        self.set_pc(run.start_addr)
        return run

    def mem_cast(self, obj, addr):
        """
        Turn bytes from an emulated memory pointer into an object
        """
        size = obj.sizeof()
        struct_bytes = self.mem_read(addr, size)
        return self.cast(obj, struct_bytes)

    def mem_purge(self):
        """
        Unmap all memory chunks
        """
        self.purge_memory()

    def setup_user_shared_data(self):
        """
        Setup the shared user data section that is often used to share data
        between user mode and kernel mode
        """
        if self.get_arch() == _arch.ARCH_X86:
            self.mem_map(self.page_size, base=0xFFDF0000,
                         tag='emu.struct.KUSER_SHARED_DATA')
        elif self.get_arch() == _arch.ARCH_AMD64:
            self.mem_map(self.page_size, base=0xFFFFF78000000000,
                         tag='emu.struct.KUSER_SHARED_DATA')

    def resume(self, addr, count=-1):
        """
        Resume emulation at the specified address.
        """
        self.emu_eng.start(addr, timeout=self.timeout,
                           count=count)

    def start(self):
        """
        Begin emulation executing each run in the specified run queue
        """
        try:
            run = self.run_queue.pop(0)
        except IndexError:
            return

        self.run_complete = False
        self.set_hooks()
        self._set_emu_hooks()
        if self.profiler:
            self.profiler.set_start_time()
        self._exec_run(run)

        while True:
            try:
                self.curr_mod = self.get_module_from_addr(self.curr_run.start_addr)
                self.emu_eng.start(self.curr_run.start_addr, timeout=self.timeout,
                                   count=self.max_instructions)
                if self.profiler:
                    if self.profiler.get_run_time() > self.timeout:
                        self.log_error('* Timeout of %d sec(s) reached.' % (self.timeout))
            except KeyboardInterrupt:
                self.log_error('* User exited.')
                return
            except Exception as e:
                if self.exit_event and self.exit_event.is_set():
                    return
                stack_trace = traceback.format_exc()

                try:
                    mnem, op, instr = self.get_disasm(self.get_pc(), DISASM_SIZE)
                except Exception as dis_err:
                    self.log_error(str(dis_err))

                error = self.get_error_info(str(e), self.get_pc(),
                                            traceback=stack_trace)
                self.curr_run.error = error

                run = self.on_run_complete()
                if not run:
                    break
                continue
            break

        self.on_emu_complete()

    def get_current_run(self):
        """
        Get the current run that is being emulated
        """
        return self.curr_run

    def get_current_module(self):
        """
        Get the currently running module
        """
        return self.curr_mod

    def get_dropped_files(self):
        """
        Get all files written by the sample from the file manager
        """
        if self.fileman:
            return self.fileman.get_dropped_files()

    def set_hooks(self):
        """
        Reserves memory that will be used to handle events that occur
        during emulation
        """
        super(WindowsEmulator, self).set_hooks()

    def get_processes(self):
        """
        Get the current processes that exist in the emulation space
        """
        if not self.processes:
            self.init_processes(self.config_processes)
        return self.processes

    def kill_process(self, proc):
        """
        Terminate a process (i.e. remove it from the known process list)
        """
        try:
            self.processes.remove(proc)
        except ValueError:
            pass

    def get_current_thread(self):
        """
        Get the current thread that is emulating
        """
        return self.curr_thread

    def get_current_process(self):
        """
        Get the current process that is emulating
        """
        return self.curr_process

    def set_current_process(self, process):
        """
        Set the current process that is emulating
        """
        self.curr_process = process

    def set_current_thread(self, thread):
        """
        Set the current thread
        """
        self.curr_thread = thread

    def _setup_gdt(self, arch):
        """
        Set up the GDT so we can access segment registers correctly
        This will be done a little differently depending on architecture
        """

        GDT_SIZE = 0x1000
        SEG_SIZE = 0x1000
        ENTRY_SIZE = 0x8
        num_gdt_entries = 31
        fs_addr = 0
        gs_addr = 0
        gdt_addr = None

        # For a detailed explaination of whats happening here, see:
        # https://wiki.osdev.org/Global_Descriptor_Table
        # We need to init the GDT so that shellcode can accurately access
        # segment registers which is needed for TEB access in user mode

        def _make_entry(index, base, access, limit=0xFFFFF000):
            access = access | (winemu.GDT_ACCESS_BITS.PresentBit |
                               winemu.GDT_ACCESS_BITS.DirectionConformingBit)
            entry = 0xFFFF & limit
            entry |= (0xFFFFFF & base) << 16
            entry |= (0xFF & access) << 40
            entry |= (0xFF & (limit >> 16)) << 48
            entry |= (0xFF & winemu.GDT_ACCESS_BITS.ProtMode32) << 52
            entry |= (0xFF & (base >> 24)) << 56
            entry = entry.to_bytes(8, 'little')

            offset = index * ENTRY_SIZE
            self.mem_write(gdt_addr + offset, entry)

        def _create_selector(index, flags):
            return flags | (index << 3)

        gdt_addr, gdt_size = self.get_valid_ranges(GDT_SIZE)
        self.mem_map(gdt_size, base=gdt_addr, tag='emu.gdt')
        seg_addr, seg_size = self.get_valid_ranges(SEG_SIZE)
        self.mem_map(seg_size, base=seg_addr, tag='emu.segment.gdt')

        access = (winemu.GDT_ACCESS_BITS.Data | winemu.GDT_ACCESS_BITS.DataWritable |
                  winemu.GDT_ACCESS_BITS.Ring3)
        _make_entry(16, 0, access)

        access = (winemu.GDT_ACCESS_BITS.Code | winemu.GDT_ACCESS_BITS.CodeReadable |
                  winemu.GDT_ACCESS_BITS.Ring3)
        _make_entry(17, 0, access)

        access = (winemu.GDT_ACCESS_BITS.Data | winemu.GDT_ACCESS_BITS.DataWritable |
                  winemu.GDT_ACCESS_BITS.Ring0)
        _make_entry(18, 0, access)

        self.reg_write(_arch.X86_REG_GDTR, (0, gdt_addr,
                                            num_gdt_entries * ENTRY_SIZE-1, 0x0))
        selector = _create_selector(16, winemu.GDT_FLAGS.Ring3)
        self.reg_write(_arch.X86_REG_DS, selector)
        selector = _create_selector(17, winemu.GDT_FLAGS.Ring3)
        self.reg_write(_arch.X86_REG_CS, selector)
        selector = _create_selector(18, winemu.GDT_FLAGS.Ring0)
        self.reg_write(_arch.X86_REG_SS, selector)

        if _arch.ARCH_X86 == arch:
            # FS segment needed for PEB access at fs:[0x30]
            fs_addr, fs_size = self.get_valid_ranges(SEG_SIZE)
            self.mem_map(fs_size, base=fs_addr, tag='emu.segment.fs')

            access = (winemu.GDT_ACCESS_BITS.Data | winemu.GDT_ACCESS_BITS.DataWritable |
                      winemu.GDT_ACCESS_BITS.Ring3)
            _make_entry(19, fs_addr, access)

            selector = _create_selector(19,  winemu.GDT_FLAGS.Ring3)
            self.reg_write(_arch.X86_REG_FS, selector)

        elif _arch.ARCH_AMD64 == arch:
            # GS Segment needed for PEB access at gs:[0x60]
            gs_addr, gs_size = self.get_valid_ranges(SEG_SIZE)
            self.mem_map(gs_size, base=gs_addr, tag='emu.segment.gs')

            access = (winemu.GDT_ACCESS_BITS.Data | winemu.GDT_ACCESS_BITS.DataWritable |
                      winemu.GDT_ACCESS_BITS.Ring3)
            _make_entry(15, gs_addr, access, limit=SEG_SIZE)

            selector = _create_selector(15,  winemu.GDT_FLAGS.Ring3)
            self.reg_write(_arch.X86_REG_GS, selector)

        self.fs_addr = fs_addr
        self.gs_addr = gs_addr

        return fs_addr, gs_addr

    def init_peb(self, user_mods, proc=None):
        """
        Initialize the Process Environment Block
        """
        p = proc
        if not p:
            p = self.curr_process
        p.init_peb(user_mods)
        self.mem_write(self.peb_addr,
                       p.peb.address.to_bytes(self.get_ptr_size(), 'little'))
        return p.peb

    def init_teb(self, thread, peb):
        """
        Initialize the Thread Information Block
        """
        if self.get_arch() == _arch.ARCH_X86:
            thread.init_teb(self.fs_addr, peb.address)
        elif self.get_arch() == _arch.ARCH_AMD64:
            thread.init_teb(self.gs_addr, peb.address)

    def init_tls(self, thread):
        """
        Initialize implicit thread local storage. Meant to be
        called after init_teb.
        """
        ptrsz = self.get_ptr_size()
        run = self.curr_run
        module = self.get_mod_from_addr(run.start_addr)

        if module:
            modname = module.emu_path
            tokens = modname.split("\\")
            modname = tokens[len(tokens) - 1]

            # Get the virtual address of the TLS directory, which will always
            # be 9 in the data directory
            tls_dirp = module.OPTIONAL_HEADER.DATA_DIRECTORY[9].VirtualAddress
            tls_dirp += module.OPTIONAL_HEADER.ImageBase

            tls_dir = self.mem_read(tls_dirp, ptrsz)

            thread.init_tls(tls_dir, os.path.splitext(modname)[0])

        return

    def load_pe(self, path=None, data=None, imp_id=winemu.IMPORT_HOOK_ADDR):
        """
        Parse a PE that will be used during emulation. PE type and architecture
        are automatically determined.
        """

        if not data and not os.path.exists(path):
            raise WindowsEmuError('File: %s not found' % (path))

        pe = winemu.PeFile(path=path, data=data, imp_id=imp_id, imp_step=4)

        pe_type = 'unknown'
        if pe.is_driver():
            pe_type = 'driver'
        elif pe.is_dll():
            pe_type = 'dll'
        elif pe.is_exe():
            pe_type = 'exe'

        arch = 'unknown'
        if pe.arch == _arch.ARCH_AMD64:
            arch = 'x64'
        elif pe.arch == _arch.ARCH_X86:
            arch = 'x86'

        self.input = {'path': pe.path, 'sha256': pe.hash,
                      'size': pe.file_size, 'arch': arch,
                      'filetype': pe_type, 'emu_version': self.get_emu_version(),
                      'os_run': self.get_osver_string()}
        if self.profiler:
            self.profiler.add_input_metadata(self.input)
        return pe

    def map_pe(self, pe, mod_name='none', emu_path=''):
        """
        Map the specified PE into the emulation space
        """
        image_size = pe.image_size
        base = pe.base
        ranges = self.get_valid_ranges(image_size, addr=base)
        base, size = ranges
        addr = self.mem_map(size, base=base, tag='emu.module.%s' % (mod_name))
        self.modules.append((pe, ranges, emu_path))

        return addr

    def get_sys_modules(self):
        """
        Get the system modules (e.g. drivers) that are loaded in the emulator
        """
        # Generate the decoy module list
        if not self.sys_modules:
            self.sys_modules = \
                self.init_sys_modules(self.config_system_modules)
        return self.sys_modules

    def get_user_modules(self):
        """
        Get the user modules (e.g. dlls) that are loaded in the emulator
        """
        # Generate the decoy user module list
        if not self.user_modules:
            self.user_modules = \
                self.init_user_modules(self.config_user_modules)
        return self.user_modules

    def get_mod_from_addr(self, addr):
        """
        Get the module (if any) that corresponds to the supplied address
        """
        if self.curr_mod:
            end = self.curr_mod.get_base() + self.curr_mod.get_image_size()
            if addr >= self.curr_mod.get_base() and addr <= end:
                return self.curr_mod

        # First check if this addr belongs to mapped decoy image we know about
        sys_mods = self.get_sys_modules()
        for m in sys_mods:
            if addr >= m.get_base() and addr < m.get_base() + m.image_size:
                return m

        user_mods = self.get_user_modules()
        for m in user_mods:
            if addr >= m.get_base() and addr < m.get_base() + m.image_size:
                return m

    def get_system_root(self):
        """
        Get the path of the "SYSTEMROOT" environment variable
        """
        sysroot = self.env.get('systemroot', 'C:\\WINDOWS\\system32')
        if not sysroot.endswith('\\'):
            sysroot += '\\'
        return sysroot

    def get_windows_dir(self):
        """
        Get the path of the "WINDIR" environment variable
        """
        sysroot = self.env.get('windir', 'C:\\WINDOWS')
        if not sysroot.endswith('\\'):
            sysroot += '\\'
        return sysroot

    def get_cd(self):
        """
        Get the path of the current directory
        """
        if not self.cd:
            self.cd = self.env.get('cd', 'C:\\WINDOWS\\system32')
            if not self.cd.endswith('\\'):
                self.cd += '\\'
        return self.cd

    def set_cd(self, cd):
        """
        Sets the current directory path
        """
        self.cd = cd

    def get_env(self):
        return self.env

    def set_env(self, var, val):
        return self.env.update({var.lower(): val})

    def get_os_version(self):
        return self.osversion

    def get_object_from_addr(self, addr):
        return self.om.get_object_from_addr(addr)

    def get_object_from_id(self, id):
        return self.om.get_object_from_id(id)

    def get_object_from_name(self, name):
        return self.om.get_object_from_name(name)

    def get_object_from_handle(self, handle):
        obj = self.om.get_object_from_handle(handle)
        if obj:
            return obj
        obj = self.fileman.get_object_from_handle(handle)
        if obj:
            return obj

    def get_object_handle(self, obj):
        obj = self.om.objects.get(obj.address)
        if obj:
            return self.om.get_handle(obj)

    def add_object(self, obj):
        self.om.add_object(obj)

    def search_path(self, file_name):
        # For now, return the current directory, add emulated path walking later
        if '\\' in file_name:
            return file_name
        fp = self.get_cd()
        if not fp.endswith('\\'):
            fp += '\\'
        return fp + file_name

    def new_object(self, otype):
        return self.om.new_object(otype)

    def create_process(self, path=None, cmdline=None, image=None, child=False):
        """
        Create a process object that will exist in the emulator
        """
        if not path and cmdline:
            path = cmdline

        # See if we are trying to create a process based off a file
        # inside the object manager and serve that
        # Setting posix to false makes shlex not treat '\' as an
        # escape character, but if each token was surrounded with
        # quotes, those are kept, so we have to delete them
        file_path = shlex.split(path, posix=False)[0]

        if file_path[0] == '\"' and file_path[len(file_path) - 1] == '\"':
            file_path = file_path[1:-1]

        p = self.om.new_object(objman.Process)

        mod_data = self.get_module_data_from_emu_file(file_path)

        if mod_data:
            # We'll create a PE out of this when we go to execute it
            p.pe_data = mod_data
        else:
            new_mod = self.init_module(name=file_path, emu_path=path)
            self.map_decoy(new_mod)
            p.pe = new_mod

        p.path = file_path
        p.cmdline = cmdline

        # Create a thread object for the new process
        t = self.om.new_object(objman.Thread)
        t.process = p
        t.tid = self.om.new_id()

        p.threads.append(t)

        if child:
            self.child_processes.append(p)
        else:
            self.processes.append(p)

        return p

    def create_thread(self, addr, ctx, proc_obj, thread_type='thread', is_suspended=False):
        """
        Create a thread object that will exist in the emulator
        """
        if len(self.run_queue) >= self.max_runs:
            return 0, None

        thread = self.om.new_object(objman.Thread)
        thread.process = proc_obj
        hnd = self.om.get_handle(thread)

        run = Run()
        run.type = thread_type
        run.start_addr = addr
        run.instr_cnt = 0
        run.args = (ctx,)
        run.process_context = proc_obj
        run.thread = thread

        if not is_suspended:
            self.run_queue.append(run)
        else:
            self.suspended_runs.append(run)

        # Returns handle
        return hnd, thread

    def resume_thread(self, thread):
        """
        Resume a previously suspended thread
        """
        for r in self.suspended_runs:
            if r.thread == thread:
                _run = self.suspended_runs.pop(self.suspended_runs.index(r))
                self.run_queue.append(_run)
                return True
        return False

    def get_dyn_imports(self):
        return self.dyn_imps

    def get_process_peb(self, process):
        return process.peb

    def add_callback(self, mod_name, func_name):
        """
        Adds a callback to the emulation callback list. A "callback" in this
        context refers to a function that in not imported statically or dynamically.

        For example, a pointer that is set in a function table
        (e.g. PsSetCreateProcessNotifyRoutine).
        """
        for addr, mod, fn in self.callbacks:
            if mod_name == mod and func_name == fn:
                return addr

        if not self.callbacks:
            curr_idx = winemu.EMU_CALLBACK_RESERVE
            self.callbacks.append((curr_idx, mod_name, func_name))
        else:
            curr_idx = self.callbacks[-1][0]
            curr_idx += 1
            self.callbacks.append((curr_idx, mod_name, func_name))

        return curr_idx

    def get_proc(self, mod_name, func_name):
        """
        Get a pointer for a supplied function name, similar to how the
        "GetProcAddress" API functions.
        """
        for addr, mod, fn in self.dyn_imps:
            if mod_name == mod and func_name == fn:
                return addr

        if not self.dyn_imps:
            curr_idx = winemu.DYM_IMP_RESERVE
            self.dyn_imps.append((curr_idx, mod_name, func_name))
        else:
            curr_idx = self.dyn_imps[-1][0]
            curr_idx += 1
            self.dyn_imps.append((curr_idx, mod_name, func_name))

        return curr_idx

    def handle_import_data(self, mod_name, sym, data_ptr=0):
        """
        Data that is imported (e.g. KeTickCount) is handled with a initializer function.
        Call it here if there is a handler for the imported variable.
        """
        module, func = self.api.get_data_export_handler(mod_name, sym)
        if not func:
            module, func = self.api.get_export_func_handler(mod_name, sym)
            if not func:
                return None

            proc_addr = self.get_proc(mod_name, sym)
            return proc_addr

        data_addr = self.api.call_data_func(module, func, data_ptr)
        return data_addr

    def _handle_invalid_fetch(self, emu, address, size, value, ctx):
        """
        Called when an attempt to emulate an instruction from an invalid address
        """
        if address == self.return_hook or address == self.exit_hook:
            self._unset_emu_hooks()
            return True

        if not self.curr_mod:
            self.curr_mod = self.get_module_from_addr(self.get_pc())

        if self.curr_mod:
            impfunc = self.curr_mod.import_table.get(address)
            if impfunc:
                mod_name, func_name = impfunc
                self.handle_import_func(mod_name, func_name)
                self._unset_emu_hooks()
                return True

        # Is the address a func ptr resolved at runtime?
        for addr, mod, fn in self.dyn_imps:
            if addr == address:
                self.handle_import_func(mod, fn)
                self._unset_emu_hooks()
                return True

        # Is the address a callback func ptr?
        for addr, mod, fn in self.callbacks:
            if addr == address:
                self.handle_import_func(mod, fn)
                self._unset_emu_hooks()
                return True

        # Are there any SEH handlers registered?
        if self.dispatch_handlers:
            rv = self.dispatch_seh(ddk.STATUS_ACCESS_VIOLATION, address)
            if rv:
                return True

        fakeout = address & 0xFFFFFFFFFFFFF000
        self.mem_map(self.page_size, base=fakeout)

        error = self.get_error_info('invalid_fetch', address)
        self.curr_run.error = error
        self.tmp_maps.append((fakeout, self.page_size))
        self.on_run_complete()
        return True

    def get_error_info(self, desc, address, traceback=None):
        """
        Collect emulator state information in the event of an error
        """
        run = self.get_current_run()
        pc = self.get_pc()
        error = {}
        self.log_error('0x%x: %s: Caught error: %s' % (pc, run.type, desc))
        error['type'] = desc
        error['pc'] = hex(pc)
        error['address'] = hex(address)
        try:
            mnem, op, instr = self.get_disasm(pc, DISASM_SIZE)
        except Exception as e:
            self.log_error(str(e))
            instr = 'disasm_failed'
        error['instr'] = instr
        error['regs'] = self.get_register_state()
        error['stack'] = self.get_stack_trace()

        if traceback:
            error['traceback'] = traceback

        return error

    def normalize_import_miss(self, dll, name):
        """
        This function attempts to fold as many function handlers together as possible.
        For example, ntdll functions will be handled by the ntoskrnl handlers, multiple versions
        of the C runtime are folded together, and Zw/Nt functions use the same handler.
        """
        alt_imp_api = ''
        alt_imp_dll = ''
        mod, func_attrs = None, None

        # Handle ANSI vs UNICODE functions
        if name.endswith('A') or name.endswith('W'):
            alt_imp_api = name[:-1]

        # Handle Zw*/Nt* function overlap
        if dll.lower().startswith('ntoskrnl'):
            if name.startswith('Zw'):
                alt_imp_api = 'Nt%s' % (name[2:])
            elif name.startswith('Nt'):
                name = 'Zw' + name[2:]
                alt_imp_api = 'Zw%s' % (name[2:])

        alt_imp_dll = winemu.normalize_dll_name(dll)

        # Bridge ntdll funcs to ntoskrnl if supported
        if dll.lower().startswith('ntdll'):
            alt_imp_dll = 'ntoskrnl'
            mod, func_attrs = self.api.get_export_func_handler(alt_imp_dll,
                                                               name)
            if not func_attrs:
                if name.startswith('Zw'):
                    alt_imp_api = 'Nt%s' % (name[2:])
                elif name.startswith('Nt'):
                    name = 'Zw' + name[2:]
                    alt_imp_api = 'Zw%s' % (name[2:])
                mod, func_attrs = self.api.get_export_func_handler(alt_imp_dll,
                                                                   alt_imp_api)
            return mod, func_attrs

        if alt_imp_api:
            mod, func_attrs = self.api.get_export_func_handler(dll,
                                                               alt_imp_api)
        elif alt_imp_dll:
            mod, func_attrs = self.api.get_export_func_handler(alt_imp_dll,
                                                               name)
        return mod, func_attrs

    def read_unicode_string(self, addr):
        """
        Read string data from a UNICODE_STRING object located at the specified address
        """
        us = windef.UNICODE_STRING(self.get_ptr_size())
        us = self.mem_cast(us, addr)

        string = self.read_mem_string(us.Buffer, width=2)
        return string

    def log_api(self, pc, imp_api, rv, argv):
        call_str = '%s(' % (imp_api)
        for arg in argv:
            if isinstance(arg, int):
                call_str += '0x%x' % (arg)
            elif isinstance(arg, str):
                call_str += '\"%s\"' % (arg.replace("\n", "\\n"))
            elif isinstance(arg, bytes):
                call_str += '\"%s\"' % (arg)
            call_str += ', '
        if call_str.endswith(', '):
            call_str = call_str[:-2]
        call_str += ')'

        _rv = rv
        if _rv is not None:
            _rv = hex(rv)
        self.log_info('%s: %s -> %s' % (hex(pc), repr(call_str), _rv))
        if self.profiler:
            # Log the API args and return value
            self.profiler.log_api(self.curr_run, pc, imp_api, rv, argv)

    def handle_import_func(self, dll, name):
        """
        Forward imported functions to the corresponding handler (if any).
        """
        imp_api = '%s.%s' % (dll, name)
        oret = self.get_ret_address()
        mod, func_attrs = self.api.get_export_func_handler(dll, name)
        if not func_attrs:
            mod, func_attrs = self.normalize_import_miss(dll, name)

        if func_attrs:
            handler_name, func, argc, conv, ordinal = func_attrs

            if name.startswith('ordinal_'):
                name = handler_name

            argv = self.get_func_argv(conv, argc)
            imp_api = '%s.%s' % (dll, name)
            default_ctx = {'func_name': imp_api}

            self.hammer.handle_import_func(imp_api, conv, argc)
            hooks = self.get_api_hooks(dll, name)
            if hooks:
                from types import MethodType
                hooked_func = MethodType(func, mod)
                orig = lambda args: hooked_func(self, args, default_ctx) # noqa
                for hook in hooks:
                    # each hook is called with the arguments, and only the last return value is
                    # considered
                    rv = hook.cb(self, imp_api, orig, argv)
            else:
                try:
                    rv = self.api.call_api_func(mod, func, argv, ctx=default_ctx)
                except Exception as e:
                    self.log_exception('0x%x: Error while calling API handler for %s:' %
                                       (oret, imp_api))
                    error = self.get_error_info(str(e), self.get_pc(),
                                                traceback=traceback.format_exc())
                    self.curr_run.error = error
                    self.on_run_complete()
                    return

            ret = self.get_ret_address()
            mm = self.get_address_map(ret)

            # Is this function being called from a dynamcially allocated memory segment?
            if mm and 'virtualalloc' in mm.get_tag().lower():
                self._dynamic_code_cb(self, ret, 0, {})

            # Log the API args and return value
            self.log_api(oret, imp_api, rv, argv)

            if not self.run_complete and ret == oret:
                self.do_call_return(argc, ret, rv, conv=conv)

        else:
            # See if a user defined a hook for this unsupported function
            hooks = self.get_api_hooks(dll, name)
            if hooks:
                # Since the function is unsupported, just call the most accurate defined hook
                hook = hooks[0]
                imp_api = '%s.%s' % (dll, name)

                if hook.call_conv is None:
                    hook.call_conv = _arch.CALL_CONV_STDCALL

                argv = self.get_func_argv(hook.call_conv, hook.argc)
                self.hammer.handle_import_func(imp_api, hook.call_conv, hook.argc)
                rv = hook.cb(self, imp_api, None, argv)
                ret = self.get_ret_address()
                self.log_api(ret, imp_api, rv, argv)
                self.do_call_return(hook.argc, ret, rv, conv=hook.call_conv)
                return
            elif self.functions_always_exist:
                imp_api = '%s.%s' % (dll, name)
                conv = _arch.CALL_CONV_STDCALL
                argc = 4
                argv = self.get_func_argv(conv, argc)
                rv = 1
                ret = self.get_ret_address()
                self.log_api(ret, imp_api, rv, argv)
                self.do_call_return(argc, ret, rv, conv=conv)
                return

            run = self.get_current_run()
            error = self.get_error_info('unsupported_api', self.get_pc())
            self.log_error("Unsupported API: %s (ret: 0x%x)" % (imp_api, oret))
            error['api_name'] = imp_api
            self.curr_run.error = error
            self.on_run_complete()

        run = self.get_current_run()
        if run and run.get_api_count() > self.max_api_count:
            self.log_info("* Maximum number of API calls reached. Stopping current run.")
            run.error['type'] = 'max_api_count'
            run.error['count'] = self.max_api_count
            run.error['pc'] = hex(self.get_pc())
            run.error['last_api'] = imp_api
            self.on_run_complete()

    def _hook_mem_unmapped(self, emu, access, address, size, value, ctx):
        """
        High level function used to catch all invalid memory accesses that occur during
        emulation
        """
        try:
            access = self.emu_eng.mem_access.get(access)
            self.prev_pc = self.get_pc()

            if not self.tmp_code_hook and not self.mem_tracing_enabled:
                self.tmp_code_hook = self.add_code_hook(cb=self._hook_code)

            self.enable_code_hook()

            if access == common.INVALID_MEM_EXEC:

                if address == winemu.SEH_RETURN_ADDR:
                    self.continue_seh()
                    self._unset_emu_hooks()
                    return True
                elif address == winemu.API_CALLBACK_HANDLER_ADDR:
                    run = self.get_current_run()
                    if run.api_callbacks:
                        pc, orig_func, args = run.api_callbacks.pop(0)
                        self.do_call_return(len(args), pc)
                        self._unset_emu_hooks()
                    return True
                return self._handle_invalid_fetch(emu, address, size,
                                                  value, ctx)

            elif access == common.INVALID_MEM_READ:
                return self._handle_invalid_read(emu, address, size,
                                                 value, ctx)

            elif access == common.INVAL_PERM_MEM_EXEC:
                return self._handle_prot_fetch(emu, address, size,
                                               value, ctx)
            elif access == common.INVALID_MEM_WRITE:

                fakeout = address & 0xFFFFFFFFFFFFF000
                self.mem_map(self.page_size, base=fakeout)
                self.tmp_maps.append((fakeout, self.page_size))

                return self._handle_invalid_write(emu, address, size,
                                                  value, ctx)
            elif access == common.INVAL_PERM_MEM_WRITE:
                return self._handle_prot_write(emu, address, size,
                                               value, ctx)
        except Exception as e:
            self.log_exception('Invalid memory exception')
            error = self.get_error_info(str(e), self.get_pc(),
                                        traceback=traceback.format_exc())
            self.curr_run.error = error
            self.on_emu_complete()
            return False

    def _handle_prot_write(self, emu, address, size, value, ctx):

        fakeout = address & 0xFFFFFFFFFFFFF000
        self.mem_map(self.page_size, base=fakeout)

        error = self.get_error_info('invalid_protect_write', address)
        self.curr_run.error = error

        self.tmp_maps.append((fakeout, self.page_size))
        self.on_run_complete()
        return True

    def restart_run(self, run):
        """
        Restart the current run
        """
        run.instr_cnt = 0
        self.set_pc(run.start_addr)

    def get_symbol_from_address(self, address):
        """
        If the supplied address is related to a known symbol, look it up here
        """
        symbol = None
        sym = self.symbols.get(address)
        if sym:
            symbol = '%s.%s' % sym
        return symbol

    def _hook_mem_read(self, emu, access, address, size, value, ctx):
        """
        Hook each memory read event that occurs. This hook is used to lookup symbols and modules
        that are read from during emulation.
        """

        try:
            symbol = self.get_symbol_from_address(address)

            if symbol:
                mod = self.get_mod_from_addr(address)
                if not mod.is_decoy():
                    mac = self.curr_run.sym_access.get(address)
                    if not mac:
                        mac = MemAccess(sym=symbol)
                    mac.reads += 1
                    self.curr_run.sym_access.update({address: mac})
                else:
                    gdata = self.global_data.get(address)
                    ptr = 0
                    if gdata:
                        symbol, ptr = gdata

                    if not ptr:
                        mn, fn = symbol.split('.')[:2]
                        data_ptr = self.handle_import_data(mn, fn)
                        if data_ptr:
                            pc = self.get_pc()
                            self.impdata_queue.append(((pc, address, symbol, data_ptr)))
                            self.set_pc(pc)
                        mac = self.curr_run.sym_access.get(address)
                        if not mac:
                            mac = MemAccess(sym=symbol)
                        mac.reads += 1
                        self.curr_run.sym_access.update({address: mac})
                        self.enable_code_hook()
                        return True

            for read_access in self.curr_run.read_cache:
                if read_access.base <= address <= (read_access.base + read_access.size) - 1:
                    read_access.reads += 1
                    return True

            mmap = self.get_address_map(address)
            if not mmap:
                return False

            maccess = self.curr_run.mem_access.get(mmap)
            if not maccess:
                maccess = MemAccess(base=mmap.base, size=mmap.size)
            self.curr_run.read_cache.appendleft(maccess)
            self.curr_run.mem_access.update({mmap: maccess})
            maccess.reads += 1

            return True
        except Exception as e:
            self.log_exception('Exception during memory read')
            error = self.get_error_info(str((type(e).__name__)), self.get_pc(),
                                        traceback=traceback.format_exc())
            self.curr_run.error = error
            self.on_emu_complete()
            return False

    def _hook_mem_write(self, emu, access, address, size, value, ctx):
        """
        Hook each memory write event that occurs. This hook is used to track memory modifications
        to interesting memory locations.
        """
        try:

            symbol = self.get_symbol_from_address(address)
            if symbol:
                mac = self.curr_run.sym_access.get(address)
                if not mac:
                    mac = MemAccess(sym=symbol)
                mac.writes += 1
                self.curr_run.sym_access.update({address: mac})

            for write_access in self.curr_run.write_cache:
                if write_access.base <= address <= (write_access.base + write_access.size) - 1:
                    write_access.writes += 1
                    return True

            mmap = self.get_address_map(address)
            if not mmap:
                return False

            maccess = self.curr_run.mem_access.get(mmap)
            if not maccess:
                maccess = MemAccess(base=mmap.base, size=mmap.size)
            self.curr_run.write_cache.appendleft(maccess)
            self.curr_run.mem_access.update({mmap: maccess})
            maccess.writes += 1

            return True

        except Exception as e:
            self.log_exception('Exception during memory write')
            error = self.get_error_info(str((type(e).__name__)), self.get_pc(),
                                        traceback=traceback.format_exc())
            self.curr_run.error = error
            self.on_emu_complete()
            return False

    def _handle_invalid_read(self, emu, address, size, value, ctx):
        """
        Hook each invalid memory read event that occurs.
        """
        # Check if the sample is trying to read from another module
        decoy = self.get_mod_from_addr(address)
        if decoy:
            return self.map_decoy(decoy)

        if address >= winemu.EMU_RESERVED and \
           address <= (winemu.EMU_RESERVED + winemu.EMU_RESERVE_SIZE):
            self._unset_emu_hooks()
            return True

        if self.dispatch_handlers:
            rv = self.dispatch_seh(ddk.STATUS_ACCESS_VIOLATION, address)
            if rv:
                return True
        fakeout = address & 0xFFFFFFFFFFFFF000
        self.mem_map(self.page_size, base=fakeout)

        error = self.get_error_info('invalid_read', address)
        self.curr_run.error = error

        # Let the next run know to remove this map since its
        # technically invalid
        self.tmp_maps.append((fakeout, self.page_size))
        self.on_run_complete()
        return True

    def _handle_prot_fetch(self, emu, address, size, value, ctx):
        """
        Called when non-executable code is emulated
        """
        # Get the symbol that the sample was trying to execute
        symbol = self.get_symbol_from_address(address)
        if not symbol:
            return True

        mac = self.curr_run.sym_access.get(address)
        if not mac:
            mac = MemAccess(sym=symbol)
        mac.execs += 1
        self.curr_run.sym_access.update({address: mac})

        mod_name, fn = symbol.split('.')

        self.handle_import_func(mod_name, fn)
        return True

    def _handle_invalid_write(self, emu, address, size, value, ctx):
        """
        Called when non-writable address is written to
        """
        # ignore patches to APIs
        if address >= winemu.EMU_RESERVED and \
                address <= (winemu.EMU_RESERVED + winemu.EMU_RESERVE_SIZE):
            return True

        if self.dispatch_handlers:
            rv = self.dispatch_seh(ddk.STATUS_ACCESS_VIOLATION, address)
            if rv:
                return True

        fakeout = address & 0xFFFFFFFFFFFFF000
        self.mem_map(self.page_size, base=fakeout)

        error = self.get_error_info('invalid_write', address)
        self.curr_run.error = error

        self.tmp_maps.append((fakeout, self.page_size))
        self.on_run_complete()
        return True

    def _hook_code(self, emu, addr, size, ctx):
        """
        Hook called before every emulated instruction. Ideally we want to
        stay out of this hook as much as possible for speed considerations.
        """

        if self.debug:
            x = self.get_disasm(addr, size)[2]
            print('0x%x: %s, edi=0x%x : esi=0x%x : ebp=0x%x : eax=0x%x' % (addr, x, self.reg_read('edi'), self.reg_read('esi'), self.reg_read('ebp'), self.reg_read('eax'))) # noqa
        try:
            if self.curr_exception_code != 0:
                self.dispatch_seh(self.curr_exception_code)
                self.curr_exception_code = 0
                self.disable_code_hook()
                return True

            if self.restart_curr_run:
                self.set_pc(self.curr_run.start_addr)
                self.restart_curr_run = False
                return False

            if addr == self.return_hook or self.run_complete:
                self.on_run_complete()
                return False

            if self.tmp_maps:
                for base, size in self.tmp_maps:
                    try:
                        self.mem_unmap(base, size)
                    except Exception:
                        self.disable_code_hook()
                        return True
                self.tmp_maps = []

            if len(self.impdata_queue):

                imp = self.impdata_queue.pop(0)
                pc, read_addr, sym, data_ptr = imp
                if data_ptr is None:
                    return True

                self.global_data.update({read_addr: [sym, data_ptr]})
                self.mem_write(read_addr,
                               data_ptr.to_bytes(self.get_ptr_size(),
                                                 'little'))
                return True

            if not self.mem_tracing_enabled:
                # Disabling the code hook here grants a significant speed bump
                if not self.debug:
                    self.disable_code_hook()

            if self.max_instructions != -1 and self.curr_run.instr_cnt >= self.max_instructions:
                self.on_run_complete()
                return False

            self.curr_instr_size = size

        # Get the symbol that the sample was trying to execute

            symbol = self.get_symbol_from_address(addr)
            if self.mem_tracing_enabled and symbol:
                mod_name, fn = symbol.split('.')

                mac = self.curr_run.sym_access.get(addr)
                if not mac:
                    mac = MemAccess(sym=symbol)
                mac.execs += 1
                self.curr_run.sym_access.update({addr: mac})

                self.handle_import_func(mod_name, fn)
                return True

            self._set_emu_hooks()

            if not self.mem_tracing_enabled:
                return

            # Increment the instruction counter
            self.curr_run.instr_cnt += 1

            for exec_access in self.curr_run.exec_cache:
                if exec_access.base <= addr <= (exec_access.base + exec_access.size) - 1:
                    exec_access.execs += 1
                    return True

            mmap = self.get_address_map(addr)
            if not mmap:
                return False
            maccess = self.curr_run.mem_access.get(mmap)
            if not maccess:
                maccess = MemAccess(base=mmap.base, size=mmap.size)
            self.curr_run.exec_cache.appendleft(maccess)
            self.curr_run.mem_access.update({mmap: maccess})
            maccess.execs += 1

            return True

        except Exception as e:
            self.log_exception('Exception during code hook')
            error = self.get_error_info(str(e), self.get_pc(),
                                        traceback=traceback.format_exc())
            self.curr_run.error = error
            self.on_emu_complete()
            return False

    def get_native_module_path(self, mod_name=''):
        """
        Get the full filesystem path of a default decoy that is supplied by
        speakeasy
        """

        def get_fp(path, mod_name):
            path = common.normalize_package_path(path)
            files = [os.path.join(path, fn) for fn in os.listdir(path)]
            for fp in files:
                bn = os.path.basename(fp.lower())
                bn = os.path.splitext(bn)[0]
                if mod_name == bn:
                    return fp

        mod_name = mod_name.lower()
        decoy_arch_dir = {_arch.ARCH_X86: ('module_directory_x86', 'x86'),
                          _arch.ARCH_AMD64: ('module_directory_x64', 'amd64')}
        dirs = decoy_arch_dir[self.get_arch()]
        mod_dir = dirs[0]

        path = self.config_modules.get(mod_dir, '')

        fp = get_fp(path, mod_name)
        if not fp:
            path = os.path.join(os.path.dirname(__file__), os.pardir, 'winenv', 'decoys', dirs[1])
            fp = get_fp(path, mod_name)

        return fp

    def load_library(self, mod_name):
        """
        Load a new module into the emulation space if necessary
        """
        ums = self.get_user_modules()
        lib = ntpath.basename(mod_name)
        lib = os.path.splitext(lib)[0]

        for um in ums:
            base = ntpath.basename(um.get_emu_path())
            base = os.path.splitext(base)[0]
            if lib.lower() == base.lower():
                hmod = um.get_base()
                return hmod

        # If we get here, the library is not found, if configured to do so,
        # we can return a fake module in every instance
        if not self.modules_always_exist:
            return 0

        mod = self.init_module(name=lib, default_base=0x6f000000)

        ums.append(mod)

        # Add the newly loaded module to the current process's PEB module list
        proc = self.get_current_process()
        if self.get_address_map(proc.get_peb_ldr().address):
            proc.add_module_to_peb(mod)

        return mod.get_base()

    def generate_export_table(self, modname):
        '''
        Generates a PE export table that can be parsed by malware
        The export names are based on the API handlers that are currently implemented
        '''

        if not modname:
            return
        modname = modname.lower()
        mod_handler = self.api.load_api_handler(modname)
        if mod_handler:

            jit = winemu.JitPeFile(self.get_arch())

            funcs = [(f[4], f[0]) for k, f in mod_handler.funcs.items() if isinstance(k, str)]
            data_exports = [k for k, d in mod_handler.data.items() if isinstance(k, str)]
            new = funcs.copy()

            if modname == 'ntdll':
                nt_handler = self.api.load_api_handler('ntoskrnl')
                funcs = [(f[4], f[0]) for k, f in nt_handler.funcs.items() if isinstance(k, str)]
                funcs = new + funcs
                new = funcs.copy()

            if modname in ('ntdll', 'ntoskrnl'):
                for o, fn in funcs:
                    if fn.startswith('Nt'):
                        new.append((None, 'Zw' + fn[2:]))
                    elif fn.startswith('Zw'):
                        new.append((None, 'Nt' + fn[2:]))
            else:
                for o, fn in funcs:
                    new.append((None, fn + 'A'))
                    new.append((None, fn + 'W'))
            func_names = new

            func_names = [fn for o, fn in func_names]
            func_names.sort()

            exports = []
            ords = [o for o, fn in funcs if o is not None]
            if ords:
                if max(ords) > len(exports):
                    num_exports = max(ords) + 1
                else:
                    num_exports = len(exports) + 1

                exports = ['ordinal_%d' % (i) for i in range(num_exports)]

                for o, fn in funcs:
                    if o is not None:
                        exports[o-1] = fn

                [exports.append(fn) for fn in func_names if fn not in exports]

            if not exports:
                exports = func_names

            exports += data_exports
            img = jit.get_decoy_pe_image(modname, exports)
            mod = winemu.DecoyModule(data=img, is_jitted=True)

            return mod
        return None

    def init_module(self, modconf={}, name='none', emu_path='', default_base=None):
        """
        Initialize a module from a config entry
        """
        modname = modconf.get('name', name)

        mod = None

        images = modconf.get('images', [])
        default_file_path = self.get_native_module_path(mod_name=modname)
        if not images:
            if not default_file_path:
                mod = self.generate_export_table(modname)
                default_file_path = self.get_native_module_path(mod_name='default_exe')

        path = None
        for img in images:
            arch = img['arch']
            if arch == self.get_arch():
                path = self.get_native_module_path(mod_name=img['name'])

        if not path:
            path = default_file_path

        if not mod:
            mod = winemu.DecoyModule(path=path)
        base = modconf.get('base_addr', default_base)
        if isinstance(base, str):
            base = int(base, 16)

        mod.decoy_path = modconf.get('path', emu_path) or (name + '.dll')
        # Reserve memory for the module
        res, size = self.get_valid_ranges(mod.image_size,
                                          base)
        mod.decoy_base = res
        mod.name = modconf.get('name', name)
        self.mem_reserve(size, base=res, tag='emu.module.%s' % (mod.name),
                         perms=common.PERM_MEM_RW)

        if mod.decoy_path == '' and name != '':
            mod.decoy_path = self.config.get('current_dir', 'C:\\Windows\\system32') + '\\' + name

        mod.base_name = ntpath.basename(mod.decoy_path)

        return mod

    # This will create a module from a file inside Speakeasy's
    # object manager. file_path is expected to point to a valid PE
    # file, like it would on a real Windows machine
    # Returns: raw data that represents a PE file
    def get_module_data_from_emu_file(self, file_path):
        if not self.does_file_exist(file_path):
            return None

        mod_file = self.fileman.get_file_from_path(file_path)

        if not mod_file:
            return None

        # This file could have been read from, so don't mess
        # with its file offset pointer. Just get the raw bytes
        # from the BytesIO object
        return mod_file.data.getvalue()

    def init_sys_modules(self, modules_config):
        """
        Initialize kernel mode modules that may be accessed by emulated modules
        """
        sys_mods = []

        for modconf in modules_config:
            mod = self.init_module(modconf)
            sys_mods.append(mod)
        return sys_mods

    def init_user_modules(self, modules_config):
        """
        Initialize user mode modules that may be accessed by emulated modules
        """
        user_mods = []

        for modconf in modules_config:
            mod = self.init_module(modconf, default_base=0x6f000000)
            user_mods.append(mod)

        return user_mods

    def map_decoy(self, decoy):
        """
        Map a decoy PE into memory. This allows samples such as shellcode to
        parse a PE's export table while resolving exported functions
        """
        if not decoy.is_mapped:
            decoy.full_load()

            for exp in decoy.get_exports():
                if exp.name:
                    sym = exp.name
                    mod_name = decoy.get_base_name()
                    addr = exp.address
                    self.symbols.update({addr: (mod_name, sym)})
                    m, hndlr = self.api.get_data_export_handler(mod_name, sym)
                    if hndlr and not self.mem_tracing_enabled:
                        self.add_mem_read_hook(cb=self._hook_mem_read, begin=addr, end=addr)
                        self.add_mem_write_hook(cb=self._hook_mem_write, begin=addr, end=addr)

            # Map the module into the emulation address space
            mem = self.mem_map_reserve(decoy.get_base())
            if mem is None:
                base, size = self.get_valid_ranges(decoy.image_size,
                                                   addr=decoy.get_base())
                mem = self.mem_map(size, base=base, tag='emu.module.%s' % (decoy.name),
                                   perms=common.PERM_MEM_RW)

            decoy.is_mapped = True
            img = decoy.get_memory_mapped_image(base=mem)
            self.mem_write(mem, bytes(img))
            if not self.mem_tracing_enabled:
                self.add_code_hook(cb=self._module_access_hook, begin=mem, end=mem+len(img))
            decoy.base = mem
            return True

    def get_thread_context(self, thread=None):
        """
        Get the current thread CPU context
        """
        if thread:
            return thread.get_context()
        else:
            ctx = self.wintypes.CONTEXT(self.get_ptr_size())
            if self.get_arch() == _arch.ARCH_X86:
                ctx.Edi = self.reg_read(_arch.X86_REG_EDI)
                ctx.Esi = self.reg_read(_arch.X86_REG_ESI)
                ctx.Eax = self.reg_read(_arch.X86_REG_EAX)
                ctx.Ebp = self.reg_read(_arch.X86_REG_EBP)
                ctx.Edx = self.reg_read(_arch.X86_REG_EDX)
                ctx.Ecx = self.reg_read(_arch.X86_REG_ECX)
                ctx.Ebx = self.reg_read(_arch.X86_REG_EBX)
                ctx.Esp = self.reg_read(_arch.X86_REG_ESP)
                ctx.Eip = self.reg_read(_arch.X86_REG_EIP)

                ctx.EFlags = self.reg_read(_arch.X86_REG_EFLAGS)
                ctx.SegCs = self.reg_read(_arch.X86_REG_CS)
                ctx.SegSs = self.reg_read(_arch.X86_REG_SS)
                ctx.SegDs = self.reg_read(_arch.X86_REG_DS)
                ctx.SegFs = self.reg_read(_arch.X86_REG_FS)
                ctx.SegGs = self.reg_read(_arch.X86_REG_GS)
                ctx.SegEs = self.reg_read(_arch.X86_REG_ES)
            elif self.get_arch() == _arch.ARCH_AMD64:
                ctx = self.wintypes.CONTEXT64(self.get_ptr_size())
        return ctx

    def load_thread_context(self, ctx, thread=None):
        """
        Set the current thread CPU context
        """
        if self.get_arch() == _arch.ARCH_X86:
            self.reg_write(_arch.X86_REG_EDI, ctx.Edi)
            self.reg_write(_arch.X86_REG_ESI, ctx.Esi)
            self.reg_write(_arch.X86_REG_EAX, ctx.Eax)
            self.reg_write(_arch.X86_REG_EBP, ctx.Ebp)
            self.reg_write(_arch.X86_REG_EDX, ctx.Edx)
            self.reg_write(_arch.X86_REG_ECX, ctx.Ecx)
            self.reg_write(_arch.X86_REG_EBX, ctx.Ebx)
            self.reg_write(_arch.X86_REG_ESP, ctx.Esp)
            self.reg_write(_arch.X86_REG_EIP, ctx.Eip)

            self.reg_write(_arch.X86_REG_EFLAGS, ctx.EFlags)
            self.reg_write(_arch.X86_REG_CS, ctx.SegCs)
            self.reg_write(_arch.X86_REG_SS, ctx.SegSs)
            self.reg_write(_arch.X86_REG_DS, ctx.SegDs)
            self.reg_write(_arch.X86_REG_FS, ctx.SegFs)
            self.reg_write(_arch.X86_REG_GS, ctx.SegGs)
            self.reg_write(_arch.X86_REG_ES, ctx.SegEs)

        elif self.get_arch() == _arch.ARCH_AMD64:
            raise NotImplementedError()

    def _get_exception_list(self):
        """
        Retrieves the exception handler list for the current thread
        """
        thread = self.get_current_thread()
        if not thread:
            return 0
        teb = thread.get_teb()
        teb = teb.read_back()
        return teb.object.NtTib.ExceptionList

    def _dispatch_seh_x86(self, except_code):
        """
        Get the initial SEH handler when dispatching a CPU exception
        that occurs during emulation
        """

        thread = self.get_current_thread()
        if not thread:
            return False
        seh = thread.get_seh()
        exception_list = self._get_exception_list()
        ptr_size = self.get_ptr_size()

        seh.last_exception_code = except_code
        # Create the _EXCEPTION_RECORD
        record = self.wintypes.EXCEPTION_RECORD(self.get_ptr_size())
        record.ExceptionCode = except_code
        record.ExceptionFlags = 0
        record.ExceptionAddress = self.get_pc()
        record.NumberParameters = 0

        ereg = self.wintypes.EXCEPTION_REGISTRATION(self.get_ptr_size())
        if exception_list:

            entry = self.mem_cast(ereg, exception_list)
            sp = self.get_stack_ptr()

            exp_ptrs = self.wintypes.EXCEPTION_POINTERS(self.get_ptr_size())

            p_exp_ptrs = self.mem_map(exp_ptrs.sizeof(), tag='emu.struct.EXCEPTION_POINTERS')
            prec = self.mem_map(record.sizeof(), tag='emu.struct.EXCEPTION_RECORD')
            pctx = self.mem_map(record.sizeof(), tag='emu.struct.EXCEPTION_CONTEXT')

            exp_ptrs.ExceptionRecord = prec
            exp_ptrs.ContextRecord = pctx

            _ctx = self.get_thread_context()
            self.mem_write(pctx, _ctx.get_bytes())
            seh.set_context(_ctx, address=pctx)

            p_exp_ptrs_bytes = (p_exp_ptrs).to_bytes(ptr_size, 'little')

            self.mem_write(p_exp_ptrs, exp_ptrs.get_bytes())
            self.mem_write(prec, record.get_bytes())

            # Write the record to the ms_exc.exc_ptr offset
            self.mem_write(exception_list - ptr_size,
                           p_exp_ptrs_bytes)

            args = [prec, exception_list, pctx, 0]
            self.set_func_args(sp, winemu.SEH_RETURN_ADDR, *args)

            run = self.get_current_run()
            regs = self.get_register_state()

            pc = self.prev_pc
            try:
                mnem, op, instr = self.get_disasm(pc, DISASM_SIZE)
            except Exception as e:
                self.log_error(str(e))
                instr = 'disasm_failed'

            self.log_info('0x%x: Exception caught: code:0x%x, handler=0x%x, instr=\"%s\"'
                          % (pc, except_code, entry.Handler, instr))

            run.handled_exceptions.append({"pc": hex(pc),
                                           "instr": instr,
                                           "exception_code": hex(except_code),
                                           "handler_address": hex(entry.Handler),
                                           "registers": regs,
                                           })

            # EBX clobber, -1 is what I observed inside a VM
            self.reg_write(_arch.X86_REG_EBX, 0xffffffff)
            self.set_pc(entry.Handler)
            return True
        return False

    def get_reserved_ranges(self):
        """
        Get the allocated memory ranges that the emulator reserves
        """
        return (winemu.EMU_RESERVED, winemu.EMU_RESERVED_END)

    def _continue_seh_x86(self):
        """
        Get the next exception handler while processing SEH
        """
        thread = self.get_current_thread()
        seh = thread.get_seh()
        sp = self.get_stack_ptr()
        ret_val = self.get_return_val()

        if seh.handler_ret_val is None:
            seh.handler_ret_val = ret_val

        ctx = seh.get_context()

        if seh.context_address:
            ctx = self.mem_cast(ctx, seh.context_address)

        # Always restore thread context, is it correct to always
        # do this?
        self.load_thread_context(ctx)

        for frame in seh.get_frames():
            if not frame.searched:
                seh.set_current_frame(frame)
                scope_record = frame.scope_records[0]
                if (not scope_record.filter_called and
                   scope_record.record.FilterFunc):
                    self.set_func_args(sp, winemu.SEH_RETURN_ADDR)
                    self.set_pc(scope_record.record.FilterFunc)
                    seh.set_last_func(scope_record.record.FilterFunc)
                    scope_record.filter_called = True
                    return

                if (windef.EXCEPTION_EXECUTE_HANDLER == ret_val or
                   scope_record.record.FilterFunc == 0 or
                   scope_record.record.FilterFunc == 0xFFFFFFFF):
                    if not scope_record.handler_called:
                        # If no filter was provided, this is a finally block
                        self.set_pc(scope_record.record.HandlerAddress)
                        seh.set_last_func(scope_record.record.HandlerAddress)
                        scope_record.handler_called = True
                        return
                elif windef.EXCEPTION_CONTINUE_EXECUTION == ret_val:
                    ctx = seh.get_context()
                    if seh.context_address:
                        _ctx = self.mem_cast(ctx, seh.context_address)
                    self.load_thread_context(_ctx)
                    self.set_pc(ctx.Eip)
                    return

                elif windef.EXCEPTION_CONTINUE_SEARCH == ret_val:
                    pass

                frame.searched = True

        if windef.EXCEPTION_CONTINUE_SEARCH == ret_val and not len(seh.get_frames()):
            ctx = seh.get_context()
            self.set_pc(ctx.Eip)
            return

        self.run_complete = True

    def dispatch_seh(self, except_code, faulting_address=None):
        rv = False
        if self.get_arch() == _arch.ARCH_X86:
            rv = self._dispatch_seh_x86(except_code)
        if not rv and self.unhandled_exception_filter:
            # Create the _EXCEPTION_RECORD
            record = self.wintypes.EXCEPTION_RECORD(self.get_ptr_size())
            record.ExceptionCode = except_code
            record.ExceptionFlags = 0
            record.ExceptionAddress = self.get_pc()
            record.NumberParameters = 0

            exp_ptrs = self.wintypes.EXCEPTION_POINTERS(self.get_ptr_size())
            p_exp_ptrs = self.mem_map(exp_ptrs.sizeof(), tag='emu.struct.EXCEPTION_POINTERS')
            prec = self.mem_map(record.sizeof(), tag='emu.struct.EXCEPTION_RECORD')
            pctx = self.mem_map(record.sizeof(), tag='emu.struct.EXCEPTION_CONTEXT')

            exp_ptrs.ExceptionRecord = prec
            exp_ptrs.ContextRecord = pctx

            self.mem_write(p_exp_ptrs, exp_ptrs.get_bytes())
            self.mem_write(prec, record.get_bytes())

            sp = self.get_stack_ptr()
            args = [p_exp_ptrs]
            self.set_func_args(sp, winemu.EMU_RETURN_ADDR, *args)
            self.set_pc(self.unhandled_exception_filter)
            self.unhandled_exception_filter = 0
            if faulting_address:
                fakeout = faulting_address & 0xFFFFFFFFFFFFF000
                self.mem_map(self.page_size, base=fakeout)
                self.tmp_maps.append((fakeout, self.page_size))
            rv = True
        return rv

    def continue_seh(self):
        if self.get_arch() == _arch.ARCH_X86:
            self._continue_seh_x86()

    def create_event(self, name=''):
        """
        Create a kernel event object
        """
        evt = self.new_object(objman.Event)
        evt.name = name
        hnd = self.om.get_handle(evt)
        return hnd, evt

    def dec_ref(self, obj):
        """
        Dereference an object
        """
        return self.om.dec_ref(obj)

    def create_mutant(self, name=''):
        """
        Create a kernel mutant object
        """
        if name == 0:
            name = ''
        mtx = self.new_object(objman.Mutant)
        mtx.name = name
        hnd = self.om.get_handle(mtx)
        return hnd, mtx

    def _hook_interrupt(self, emu, intnum, ctx=[]):
        """
        Called when software interrupts occur
        """
        def _tmp_hook(emu, addr, size, ctx):
            ret = self.pop_stack()
            self.set_pc(ret)
            hook_obj = ctx.pop(0)
            hook_obj.disable()

        exception_list = self._get_exception_list()
        if exception_list and self.dispatch_handlers:
            # Catch software breakpoint interrupts
            if intnum == 3 or intnum == 0x2d:
                self.curr_exception_code = ddk.STATUS_BREAKPOINT
                self.prev_pc = self.get_pc()
                self.enable_code_hook()
                return True
            # Catch single step exceptions
            elif intnum == 1:
                self.curr_exception_code = ddk.STATUS_SINGLE_STEP
                self.enable_code_hook()
                self.prev_pc = self.get_pc()
                eflags = self.reg_read(_arch.X86_REG_EFLAGS)
                # Remove the trap flag
                eflags &= 0xFFFFFEFF
                self.reg_write(_arch.X86_REG_EFLAGS, eflags)
                return True

        # Handle __fastfail interrupt introduced in Windows 8
        if intnum == 0x29:
            ecx = self.reg_read(_arch.X86_REG_ECX)
            # Cookie security init failed, just return since we are in __security_init_cookie
            if ecx == 6:
                ctx.append(self.add_code_hook(cb=_tmp_hook, ctx=ctx))
                return True

        pc = self.get_pc()
        self.log_error('0x%x: Unhandled interrupt: intnum=0x%x' % (pc, intnum))
        error = self.get_error_info('unhandled_interrupt', pc)
        error.update({'interrupt_num': intnum})
        self.curr_run.error = error

        self.restart_curr_run = True
        self.on_run_complete()
        return False



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

# EXTENDED_NAME_FORMAT
NameUnknown = 0
NameFullyQualifiedDN = 1
NameSamCompatible = 2
NameDisplay = 3
NameUniqueId = 6
NameCanonical = 7
NameUserPrincipal = 8
NameCanonicalEx = 9
NameServicePrincipal = 0xA
NameDnsDomain = 0xC
NameGivenName = 0xD
NameSurname = 0xE

SEC_E_INVALID_HANDLE = 0x80090301


def get_define(define, prefix=''):
    for k, v in globals().items():
        if not isinstance(v, int) or v != define:
            continue
        if prefix:
            if k.startswith(prefix):
                return k
        else:
            return k



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

from speakeasy.struct import Enum

WINDOWS_CONSOLE = 3

# IRQL Levels
PASSIVE_LEVEL = 0             # Passive release level
LOW_LEVEL = 0                 # Lowest interrupt level
APC_LEVEL = 1                 # APC interrupt level
DISPATCH_LEVEL = 2            # Dispatcher level
CMCI_LEVEL = 5                # CMCI handler level
PROFILE_LEVEL = 27            # timer used for profiling.
CLOCK1_LEVEL = 28             # Interval clock 1 level - Not used on x86
CLOCK2_LEVEL = 28             # Interval clock 2 level
IPI_LEVEL = 29                # Interprocessor interrupt level
POWER_LEVEL = 30              # Power failure level
HIGH_LEVEL = 31               # Highest interrupt level

STATUS_SUCCESS = 0
STATUS_BREAKPOINT = 0x80000003
STATUS_SINGLE_STEP = 0x80000004
STATUS_UNSUCCESSFUL = 0xC0000001
STATUS_INFO_LENGTH_MISMATCH = 0xC0000004
STATUS_ACCESS_VIOLATION = 0xC0000005
STATUS_INVALID_HANDLE = 0xC0000008
STATUS_ILLEGAL_INSTRUCTION = 0xC000001D
STATUS_PRIVILEGED_INSTRUCTION = 0xC0000096
STATUS_BUFFER_TOO_SMALL = 0xC0000023
STATUS_INVALID_CID = 0xC000000B
STATUS_INVALID_PARAMETER = 0xC000000D
STATUS_PROCEDURE_NOT_FOUND = 0xC000007A
STATUS_NOT_SUPPORTED = 0xC00000BB
STATUS_OBJECT_NAME_NOT_FOUND = 0xC0000034
STATUS_DEBUGGER_INACTIVE = 0xC0000354
STATUS_BAD_COMPRESSION_BUFFER = 0xC0000242
STATUS_UNSUPPORTED_COMPRESSION = 0xC000025F
STATUS_INVALID_DEVICE_REQUEST = 0xC0000010
STATUS_OBJECT_TYPE_MISMATCH = 0xC0000024
STATUS_NOINTERFACE = 0xC00002B9
STATUS_PORT_NOT_SET = 0xC0000353

# Device Flags
DO_DIRECT_IO = 0x00000010
DO_BUFFERED_IO = 0x00000004
DO_EXCLUSIVE = 0x00000008
DO_DEVICE_INITIALIZING = 0x00000080

IRP_MJ_CREATE = 0x00
IRP_MJ_CREATE_NAMED_PIPE = 0x01
IRP_MJ_CLOSE = 0x02
IRP_MJ_READ = 0x03
IRP_MJ_WRITE = 0x04
IRP_MJ_QUERY_INFORMATION = 0x05
IRP_MJ_SET_INFORMATION = 0x06
IRP_MJ_QUERY_EA = 0x07
IRP_MJ_SET_EA = 0x08
IRP_MJ_FLUSH_BUFFERS = 0x09
IRP_MJ_QUERY_VOLUME_INFORMATION = 0x0a
IRP_MJ_SET_VOLUME_INFORMATION = 0x0b
IRP_MJ_DIRECTORY_CONTROL = 0x0c
IRP_MJ_FILE_SYSTEM_CONTROL = 0x0d
IRP_MJ_DEVICE_CONTROL = 0x0e
IRP_MJ_INTERNAL_DEVICE_CONTROL = 0x0f
IRP_MJ_SHUTDOWN = 0x10
IRP_MJ_LOCK_CONTROL = 0x11
IRP_MJ_CLEANUP = 0x12
IRP_MJ_CREATE_MAILSLOT = 0x13
IRP_MJ_QUERY_SECURITY = 0x14
IRP_MJ_SET_SECURITY = 0x15
IRP_MJ_POWER = 0x16
IRP_MJ_SYSTEM_CONTROL = 0x17
IRP_MJ_DEVICE_CHANGE = 0x18
IRP_MJ_QUERY_QUOTA = 0x19
IRP_MJ_SET_QUOTA = 0x1a
IRP_MJ_PNP = 0x1b
IRP_MJ_PNP_POWER = IRP_MJ_PNP
IRP_MJ_MAXIMUM_FUNCTION = 0x1b

COMPRESSION_FORMAT_LZNT1 = 0x2
COMPRESSION_FORMAT_XPRESS = 0x3

DELETE = 0x00010000
READ_CONTROL = 0x00020000
WRITE_DAC = 0x00040000
WRITE_OWNER = 0x00080000
SYNCHRONIZE = 0x00100000
GENERIC_READ = 0x80000000
GENERIC_WRITE = 0x40000000
GENERIC_EXECUTE = 0x20000000
GENERIC_ALL = 0x10000000

# Create Dispositions
FILE_SUPERSEDE = 0x00000000
FILE_OPEN = 0x00000001
FILE_CREATE = 0x00000002
FILE_OPEN_IF = 0x00000003
FILE_OVERWRITE = 0x00000004
FILE_OVERWRITE_IF = 0x00000005

# File specific access mask
FILE_READ_DATA = 0x0001  # file & pipe
FILE_WRITE_DATA = 0x0002  # file & pipe
FILE_APPEND_DATA = 0x0004  # file
FILE_READ_EA = 0x0008  # file & directory
FILE_WRITE_EA = 0x0010  # file & directory
FILE_EXECUTE = 0x0020  # file
FILE_DELETE_CHILD = 0x0040  # directory
FILE_READ_ATTRIBUTES = 0x0080  # all
FILE_WRITE_ATTRIBUTES = 0x0100  # all


PROCESSINFOCLASS = Enum()
PROCESSINFOCLASS.ProcessBasicInformation = 0
PROCESSINFOCLASS.ProcessDebugPort = 7
PROCESSINFOCLASS.ProcessWow64Information = 0x1a
PROCESSINFOCLASS.ProcessImageFileName = 0x1b
PROCESSINFOCLASS.ProcessBreakOnTermination = 0x1d
PROCESSINFOCLASS.ProcessDebugObjectHandle = 0x1e
PROCESSINFOCLASS.ProcessProtectionInformation = 0x3d


SYSTEM_INFORMATION_CLASS = Enum()
SYSTEM_INFORMATION_CLASS.SystemBasicInformation = 0x00
SYSTEM_INFORMATION_CLASS.SystemProcessorInformation = 0x01
SYSTEM_INFORMATION_CLASS.SystemPerformanceInformation = 0x02
SYSTEM_INFORMATION_CLASS.SystemTimeOfDayInformation = 0x03
SYSTEM_INFORMATION_CLASS.SystemPathInformation = 0x04
SYSTEM_INFORMATION_CLASS.SystemProcessInformation = 0x05
SYSTEM_INFORMATION_CLASS.SystemCallCountInformation = 0x06
SYSTEM_INFORMATION_CLASS.SystemDeviceInformation = 0x07
SYSTEM_INFORMATION_CLASS.SystemProcessorPerformanceInformation = 0x08
SYSTEM_INFORMATION_CLASS.SystemFlagsInformation = 0x09
SYSTEM_INFORMATION_CLASS.SystemCallTimeInformation = 0x0A
SYSTEM_INFORMATION_CLASS.SystemModuleInformation = 0x0B
SYSTEM_INFORMATION_CLASS.SystemKernelDebuggerInformation = 0x23


FILE_INFORMATION_CLASS = Enum()
FILE_INFORMATION_CLASS.FileDirectoryInformation = 1
FILE_INFORMATION_CLASS.FileFullDirectoryInformation = 2
FILE_INFORMATION_CLASS.FileBothDirectoryInformation = 3
FILE_INFORMATION_CLASS.FileBasicInformation = 4
FILE_INFORMATION_CLASS.FileStandardInformation = 5
FILE_INFORMATION_CLASS.FileInternalInformation = 6
FILE_INFORMATION_CLASS.FileEaInformation = 7
FILE_INFORMATION_CLASS.FileAccessInformation = 8
FILE_INFORMATION_CLASS.FileNameInformation = 9
FILE_INFORMATION_CLASS.FileRenameInformation = 10
FILE_INFORMATION_CLASS.FileLinkInformation = 11


POOL_TYPE = Enum()
POOL_TYPE.NonPagedPool = 0
POOL_TYPE.PagedPool = 1
POOL_TYPE.NonPagedPoolMustSucceed = 2
POOL_TYPE.DontUseThisType = 3
POOL_TYPE.NonPagedPoolCacheAligned = 4
POOL_TYPE.PagedPoolCacheAligned = 5
POOL_TYPE.NonPagedPoolCacheAlignedMustS = 6
POOL_TYPE.MaxPoolType = 7
POOL_TYPE.NonPagedPoolSession = 32
POOL_TYPE.PagedPoolSession = 33
POOL_TYPE.NonPagedPoolMustSucceedSession = 34
POOL_TYPE.DontUseThisTypeSession = 35
POOL_TYPE.NonPagedPoolCacheAlignedSession = 36
POOL_TYPE.PagedPoolCacheAlignedSession = 37
POOL_TYPE.NonPagedPoolCacheAlignedMustSSession = 38
POOL_TYPE.NonPagedPoolNx = 512

MODE = Enum()
MODE.KernelMode = 0
MODE.UserMode = 1
MODE.MaximumMode = 2

IMAGE_DOS_SIGNATURE = b'MZ'
PE32_BIT = 0x0100
PE32_PLUS_BIT = 0x0200


def get_flag_defines(flags, prefix=''):
    defs = []
    for k, v in globals().items():
        if isinstance(v, int):
            if v & flags:
                if prefix:
                    if k.startswith(prefix):
                        defs.append(k)
                else:
                    defs.append(k)
    return defs


def get_const_defines(const, prefix=''):
    defs = []
    for k, v in globals().items():
        if isinstance(v, int):
            if v == const:
                if prefix:
                    if k.startswith(prefix):
                        defs.append(k)
                else:
                    defs.append(k)
    return defs


def get_access_defines(flags):
    defs = []
    accesses = ('DELETE', 'READ_CONTROL', 'WRITE_DAC',
                'WRITE_OWNER', 'SYNCHRONIZE', 'GENERIC_READ',
                'GENERIC_WRITE', 'GENERIC_EXECUTE', 'GENERIC_ALL')

    for k, v in [(k, v) for k, v in globals().items() if k in accesses]:
        if isinstance(v, int):
            if v & flags:
                defs.append(k)

    return defs


def get_file_access_defines(flags):
    defs = get_flag_defines(flags, 'FILE_')
    defs = [d for d in defs if d.startswith(('FILE_READ', 'FILE_WRITE',
                                            'FILE_DELETE', 'FILE_APPEND', 'FILE_EXECUTE'))]

    return defs


def get_create_disposition(disp):
    defs = get_const_defines(disp, 'FILE_')
    defs = [d for d in defs if not d.startswith('FILE_SHARE')]

    ret = 0
    if len(defs):
        ret = defs[0]
    return ret



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

import base64

from .. import api

import speakeasy.winenv.defs.nt.ddk as ntdefs


class Bcrypt(api.ApiHandler):
    """
    Implements exported functions from bcrypt.dll
    """
    name = 'bcrypt'
    apihook = api.ApiHandler.apihook
    impdata = api.ApiHandler.impdata

    def __init__(self, emu):

        super(Bcrypt, self).__init__(emu)

        self.funcs = {}
        self.data = {}

        super(Bcrypt, self).__get_hook_attrs__(self)

    @apihook('BCryptOpenAlgorithmProvider', argc=4)
    def BCryptOpenAlgorithmProvider(self, emu, argv, ctx={}):
        """
        NTSTATUS BCryptOpenAlgorithmProvider(
          BCRYPT_ALG_HANDLE *phAlgorithm,
          LPCWSTR           pszAlgId,
          LPCWSTR           pszImplementation,
          ULONG             dwFlags
        );
        """
        phAlgorithm, pszAlgId, pszImplementation, dwFlags = argv

        algid = self.read_wide_string(pszAlgId)
        if algid:
            argv[1] = algid

        implementation = ''
        if pszImplementation:
            implementation = self.read_wide_string(pszImplementation)
            if implementation:
                argv[2] = implementation

        cm = emu.get_crypt_manager()
        hnd = cm.crypt_open(pname=implementation, ptype=algid, flags=dwFlags)
        if hnd:
            self.mem_write(phAlgorithm,
                           hnd.to_bytes(emu.get_ptr_size(), 'little'))
            argv[0] = hnd

        return ntdefs.STATUS_SUCCESS

    @apihook('BCryptImportKeyPair', argc=7)
    def BCryptImportKeyPair(self, emu, argv, ctx={}):
        """
        NTSTATUS BCryptImportKeyPair(
          BCRYPT_ALG_HANDLE hAlgorithm,
          BCRYPT_KEY_HANDLE hImportKey,
          LPCWSTR           pszBlobType,
          BCRYPT_KEY_HANDLE *phKey,
          PUCHAR            pbInput,
          ULONG             cbInput,
          ULONG             dwFlags
        );
        """
        hAlgorithm, hImportKey, pszBlobType, phKey, pbInput, cbInput, dwFlags \
            = argv

        blob_type = self.read_wide_string(pszBlobType)
        argv[2] = blob_type

        cbInput = cbInput & 0xFFFFFFFF
        blob = self.mem_read(pbInput, cbInput)
        argv[4] = base64.b64encode(blob).decode('utf-8')

        cm = emu.get_crypt_manager()
        if hAlgorithm and phKey:
            ctx = cm.crypt_get(hAlgorithm)
            hnd = ctx.import_key(blob_type=blob_type,
                                 blob=blob,
                                 blob_len=cbInput,
                                 flags=dwFlags)
            if hnd:
                self.mem_write(phKey,
                               hnd.to_bytes(emu.get_ptr_size(), 'little'))
                argv[3] = hnd

        return ntdefs.STATUS_SUCCESS

    @apihook('BCryptCloseAlgorithmProvider', argc=2)
    def BCryptCloseAlgorithmProvider(self, emu, argv, ctx={}):
        """
        NTSTATUS BCryptCloseAlgorithmProvider(
          BCRYPT_ALG_HANDLE hAlgorithm,
          ULONG             dwFlags
        );
        """
        hAlgorithm, dwFlags = argv

        cm = emu.get_crypt_manager()
        if hAlgorithm:
            cm.crypt_close(hAlgorithm)

        return ntdefs.STATUS_SUCCESS

    @apihook('BCryptGetProperty', argc=6)
    def BCryptGetProperty(self, emu, argv, ctx={}):
        """
        NTSTATUS BCryptGetProperty(
          BCRYPT_HANDLE hObject,
          LPCWSTR       pszProperty,
          PUCHAR        pbOutput,
          ULONG         cbOutput,
          ULONG         *pcbResult,
          ULONG         dwFlags
        );
        """
        hObject, pszProperty, pbOutput, cbOutput, pcbResult, dwFlags = argv

        property = self.read_wide_string(pszProperty)
        if property:
            argv[1] = property

        # TODO: implement property retrieval

        return ntdefs.STATUS_SUCCESS

    @apihook('BCryptDestroyKey', argc=1)
    def BCryptDestroyKey(self, emu, argv, ctx={}):
        """
        NTSTATUS BCryptDestroyKey(
          BCRYPT_KEY_HANDLE hKey
        );
        """
        hKey, = argv
        cm = emu.get_crypt_manager()
        for hnd, ctx in cm.ctx_handles.items():
            hnd_key = ctx.get_key(hKey)
            if hnd_key:
                ctx.delete_key(hKey)
                return ntdefs.STATUS_SUCCESS

        return ntdefs.STATUS_INVALID_HANDLE



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

import speakeasy.winenv.arch as _arch
import speakeasy.winenv.defs.windows.secur32 as sec32defs

from .. import api


class Secur32(api.ApiHandler):

    name = 'secur32'
    apihook = api.ApiHandler.apihook
    impdata = api.ApiHandler.impdata

    def __init__(self, emu):

        super(Secur32, self).__init__(emu)
        super(Secur32, self).__get_hook_attrs__(self)

    @apihook('GetUserNameEx', argc=3, conv=_arch.CALL_CONV_STDCALL)
    def GetUserNameEx(self, emu, argv, ctx={}):
        """
        BOOLEAN SEC_ENTRY GetUserNameExA(
          EXTENDED_NAME_FORMAT NameFormat,
          LPSTR                lpNameBuffer,
          PULONG               nSize
        );
        """
        NameFormat, lpNameBuffer, nSize = argv

        cw = self.get_char_width(ctx)

        name_format = sec32defs.get_define(NameFormat, prefix='Name')
        if name_format:
            argv[0] = name_format

        user = emu.get_user()
        user_name = user.get('name')
        user_name_len = len(user_name)

        argv[1] = user_name
        argv[2] = user_name_len

        self.write_mem_string(user_name, lpNameBuffer, cw)
        self.mem_write(nSize, user_name_len.to_bytes(4, 'little'))

        return 1

    @apihook('EncryptMessage', argc=4)
    def EncryptMessage(self, emu, argv, ctx={}):
        """
        SECURITY_STATUS SEC_ENTRY EncryptMessage(
        PCtxtHandle    phContext,
        unsigned long  fQOP,
        PSecBufferDesc pMessage,
        unsigned long  MessageSeqNo
        );
        """

        PCtxtHandle, fQOP, pMessage, MessageSeqNo = argv

        return sec32defs.SEC_E_INVALID_HANDLE



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

ARCH_X86 = 32
ARCH_AMD64 = 64
PAGE_SIZE = 0x1000
BITS_32 = 32
BITS_64 = 64

LSTAR = 0xC0000082

X86_REG_CS = 1001
X86_REG_DS = 1002
X86_REG_EAX = 1003
X86_REG_EBP = 1004
X86_REG_EBX = 1005
X86_REG_ECX = 1006
X86_REG_EDI = 1007
X86_REG_EDX = 1008
X86_REG_EFLAGS = 1009
X86_REG_EIP = 1010
X86_REG_EIZ = 1011
X86_REG_ESI = 1012
X86_REG_ESP = 1013

# Segment registers
X86_REG_FS = 1014
X86_REG_GS = 1015
X86_REG_ES = 1016
X86_REG_SS = 1017
X86_REG_CS = 1018
X86_REG_DS = 1019

X86_REG_MSR = 1020

# AMD64
AMD64_REG_RAX = 1021
AMD64_REG_RBP = 1022
AMD64_REG_RBX = 1023
AMD64_REG_RCX = 1024
AMD64_REG_RDI = 1025
AMD64_REG_RDX = 1026
AMD64_REG_RIP = 1027
AMD64_REG_RIZ = 1028
AMD64_REG_RSI = 1029
AMD64_REG_RSP = 1030
AMD64_REG_SIL = 1031
AMD64_REG_DIL = 1032
AMD64_REG_BPL = 1033
AMD64_REG_SPL = 1034
AMD64_REG_R8 = 1035
AMD64_REG_R9 = 1036
AMD64_REG_R10 = 1037
AMD64_REG_R11 = 1038
AMD64_REG_R12 = 1039
AMD64_REG_R13 = 1040
AMD64_REG_R14 = 1041
AMD64_REG_R15 = 1042

# Control Registers
X86_REG_CR0 = 1043
X86_REG_CR1 = 1044
X86_REG_CR2 = 1045
X86_REG_CR3 = 1046
X86_REG_CR4 = 1047
X86_REG_CR5 = 1048
X86_REG_CR6 = 1049
X86_REG_CR7 = 1050
X86_REG_CR8 = 1051

# Debug registers
X86_REG_DR0 = 1052
X86_REG_DR1 = 1053
X86_REG_DR2 = 1054
X86_REG_DR3 = 1055
X86_REG_DR4 = 1056
X86_REG_DR5 = 1057
X86_REG_DR6 = 1058
X86_REG_DR7 = 1059
X86_REG_DR8 = 1060

X86_REG_IDTR = 1061
X86_REG_GDTR = 1062

X86_REG_XMM0 = 1063
X86_REG_XMM1 = 1064
X86_REG_XMM2 = 1065
X86_REG_XMM3 = 1066
X86_REG_XMM4 = 1067

# Calling conventions
CALL_CONV_CDECL = 0
CALL_CONV_STDCALL = 1
CALL_CONV_FASTCALL = 2
CALL_CONV_FLOAT = 3
VAR_ARGS = -1

# Register string constants
REG_LOOKUP = {
    # x86 registers
    'eax': X86_REG_EAX, 'ebx': X86_REG_EBX, 'ecx': X86_REG_ECX, 'edx': X86_REG_EDX,
    'edi': X86_REG_EDI, 'esi': X86_REG_ESI, 'ebp': X86_REG_EBP, 'esp': X86_REG_ESP,
    'eip': X86_REG_EIP, 'eflags': X86_REG_EFLAGS,
    # amd64 registers
    'rax': AMD64_REG_RAX, 'rbx': AMD64_REG_RBX, 'rcx': AMD64_REG_RCX, 'rdx': AMD64_REG_RDX,
    'rdi': AMD64_REG_RDI, 'rsi': AMD64_REG_RSI, 'rsp': AMD64_REG_RSP, 'rbp': AMD64_REG_RBP,
    'r8': AMD64_REG_R8, 'r9': AMD64_REG_R9, 'r10': AMD64_REG_R10, 'r11': AMD64_REG_R11,
    'r12': AMD64_REG_R12, 'r13': AMD64_REG_R13, 'r14': AMD64_REG_R14, 'r15': AMD64_REG_R15,
    'rip': AMD64_REG_RIP
}



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

from .. import api
import shlex
import speakeasy.winenv.defs.windows.windows as windefs
import speakeasy.winenv.defs.windows.shell32 as shell32_defs


class Shell32(api.ApiHandler):

    """
    Implements exported functions from shell32.dll
    """

    name = 'shell32'
    apihook = api.ApiHandler.apihook
    impdata = api.ApiHandler.impdata

    def __init__(self, emu):

        super(Shell32, self).__init__(emu)

        self.funcs = {}
        self.data = {}
        self.window_hooks = {}
        self.handle = 0
        self.win = None
        self.curr_handle = 0x2800

        super(Shell32, self).__get_hook_attrs__(self)

    def get_handle(self):
        self.curr_handle += 4
        return self.curr_handle

    @apihook('SHCreateDirectoryEx', argc=3)
    def SHCreateDirectoryEx(self, emu, argv, ctx={}):
        '''
        int SHCreateDirectoryExA(
            HWND                      hwnd,
            LPCSTR                    pszPath,
            const SECURITY_ATTRIBUTES *psa
        );
        '''

        hwnd, pszPath, psa = argv

        cw = self.get_char_width(ctx)
        dn = ''
        if pszPath:
            dn = self.read_mem_string(pszPath, cw)
            argv[1] = dn

            self.log_file_access(dn, 'directory_create')

        return 0

    @apihook('ShellExecute', argc=6)
    def ShellExecute(self, emu, argv, ctx={}):
        '''
        HINSTANCE ShellExecuteA(
            HWND   hwnd,
            LPCSTR lpOperation,
            LPCSTR lpFile,
            LPCSTR lpParameters,
            LPCSTR lpDirectory,
            INT    nShowCmd
        );
        '''

        hwnd, lpOperation, lpFile, lpParameters, lpDirectory, nShowCmd = argv

        cw = self.get_char_width(ctx)

        fn = ''
        param = ''
        dn = ''
        if lpOperation:
            op = self.read_mem_string(lpOperation, cw)
            argv[1] = op
        if lpFile:
            fn = self.read_mem_string(lpFile, cw)
            argv[2] = fn
        if lpParameters:
            param = self.read_mem_string(lpParameters, cw)
            argv[3] = param
        if lpDirectory:
            dn = self.read_mem_string(lpDirectory, cw)
            argv[4] = dn

        if dn and fn:
            fn = '%s\\%s' % (dn, fn)

        proc = emu.create_process(path=fn, cmdline=param)
        self.log_process_event(proc, 'create')

        return 33

    @apihook('IsUserAnAdmin', argc=0)
    def IsUserAnAdmin(self, emu, argv, ctx={}):
        """
        BOOL IsUserAnAdmin();
        """
        return emu.get_user().get('is_admin', False)

    @apihook('SHGetMalloc', argc=1)
    def SHGetMalloc(self, emu, argv, ctx={}):
        """
        SHSTDAPI SHGetMalloc(
            IMalloc **ppMalloc
        );
        """
        ppMalloc, = argv

        if ppMalloc:
            ci = emu.com.get_interface(emu, emu.get_ptr_size(), 'IMalloc')
            self.mem_write(ppMalloc, ci.address.to_bytes(emu.get_ptr_size(), 'little'))
        rv = windefs.S_OK
        return rv

    @apihook('CommandLineToArgv', argc=2)
    def CommandLineToArgv(self, emu, argv, ctx={}):
        """
        LPWSTR * CommandLineToArgv(
            LPCWSTR lpCmdLine,
            int     *pNumArgs
        );
        """
        cmdline, argc = argv

        cw = self.get_char_width(ctx)
        cl = self.read_mem_string(cmdline, cw)

        ptrsize = emu.get_ptr_size()

        split = shlex.split(cl)
        nargs = len(split)

        # Get the total size we need
        size = (len(split) + 1) * ptrsize
        size += (len(cl) * cw) + (len(split) * cw)

        # Allocate the array
        buf = self.mem_alloc(size, tag='api.CommandLineToArgv')
        ptrs = buf
        strs = buf + ((len(split) + 1) * ptrsize)
        for i, p in enumerate(split):
            self.mem_write(ptrs + (i * ptrsize), strs.to_bytes(emu.get_ptr_size(), 'little'))

            p += '\x00'
            if cw == 2:
                s = p.encode('utf-16le')
            else:
                s = p.encode('utf-8')
            self.mem_write(strs, s)

            strs += len(s)

        if argc:
            self.mem_write(argc, nargs.to_bytes(4, "little"))

        return buf

    @apihook('ExtractIcon', argc=3)
    def ExtractIcon(self, emu, argv, ctx={}):
        """
        HICON ExtractIconA(
          HINSTANCE hInst,
          LPCSTR    pszExeFileName,
          UINT      nIconIndex
        );
        """

        return self.get_handle()

    @apihook('SHGetFolderPath', argc=5)
    def SHGetFolderPath(self, emu, argv, ctx={}):
        """
        HWND   hwnd,
        int    csidl,
        HANDLE hToken,
        DWORD  dwFlags,
        LPWSTR pszPath
        """
        hwnd, csidl, hToken, dwFlags, pszPath = argv
        if csidl in shell32_defs.CSIDL:
            argv[1] = shell32_defs.CSIDL[csidl]
        if csidl == 0x1a:
            # CSIDL_APPDATA
            path = "C:\\Users\\{}\\AppData\\Roaming".format(emu.get_user()['name'])
        elif csidl == 0x28:
            # csidl_profile
            path = "C:\\Users\\{}".format(emu.get_user()['name'])
        elif csidl == 0 or csidl == 0x10:
            # CSIDL_DESKTOP or CSIDL_DESKTOPDIRECTORY
            path = "C:\\Users\\{}\\Desktop".format(emu.get_user()['name'])
        elif csidl == 2:
            # CSIDL_PROGRAMS
            path = "C:\\Users\\{}\\AppData\\Roaming\\Microsoft\\Windows\\Start Menu\\Programs".format(emu.get_user()['name']) # noqa
        elif csidl == 6 or csidl == 0x1f:
            # CSIDL_FAVORITES or CSIDL_COMMON_FAVORITES
            path = "C:\\Users\\{}\\Favorites".format(emu.get_user()['name'])
        elif csidl == 7:
            # CSIDL_STARTUP
            path = "C:\\Users\\{}\\\AppData\\Roaming\\Microsoft\\Windows\\Start Menu\\Programs\\Startup".format(emu.get_user()['name']) # noqa
        elif csidl == 8:
            # CSIDL_RECENT
            path = "C:\\Users\\{}\\\AppData\\Roaming\\Microsoft\\Windows\\Recent".format(emu.get_user()['name']) # noqa
        elif csidl == 9:
            # csidl_sendto
            path = "C:\\Users\\{}\\AppData\\Roaming\\Microsoft\\Windows\\SendTo".format(emu.get_user()['name']) # noqa
        elif csidl == 0xb:
            # CSIDL_STARTMENU
            path = "C:\\Users\\{}\\AppData\\Roaming\\Microsoft\\Windows\\Start Menu".format(emu.get_user()['name']) # noqa
        elif csidl == 0x13:
            # CSIDL_NETHOOD
            path = "C:\\Users\\{}\\AppData\\Roaming\\Microsoft\\Windows\\Network Shortcuts".format(emu.get_user()['name']) # noqa
        elif csidl == 0x15:
            # CSIDL_TEMPLATES
            path = "C:\\Users\\{}\\AppData\\Roaming\\Microsoft\\Windows\\Templates".format(emu.get_user()['name']) # noqa
        elif csidl == 0x1b:
            # CSIDL_PRINTHOOD
            path = "C:\\Users\\{}\\AppData\\Roaming\\Microsoft\\Windows\\Printer Shortcuts".format(emu.get_user()['name']) # noqa
        elif csidl == 0x1c:
            # CSIDL_LOCAL_APPDATA
            path = "C:\\Users\\{}\\AppData\\Local".format(emu.get_user()['name'])
        elif csidl == 0x20:
            # CSIDL_INTERNET_CACHE
            path = "C:\\Users\\{}\\AppData\\Local\\Microsoft\\Windows\\Temporary Internet File".format(emu.get_user()['name']) # noqa
        elif csidl == 0x21:
            # CSIDL_COOKIES
            path = "C:\\Users\\{}\\AppData\\AppData\\Roaming\\Microsoft\\Windows\\Cookies".format(emu.get_user()['name']) # noqa
        elif csidl == 0x22:
            # CSIDL_HISTORY
            path = "C:\\Users\\{}\\AppData\\Local\\Microsoft\\Windows\\History".format(emu.get_user()['name']) # noqa
        elif csidl == 0x27:
            # CSIDL_MYPICTURES
            path = "C:\\Users\\{}\\Pictures".format(emu.get_user()['name'])
        elif csidl == 0x2f or csidl == 0x30:
            path = "C:\\Users\\{}\\AppData\\Roaming\\Microsoft\\Windows\\Start Menu\\Programs\\Administrative Tools".format(emu.get_user()['name']) # noqa
        elif csidl == 0x1d:
            # CSIDL_ALTSTARTUP
            path = "C:\\Users\\{}\\AppData\\Roaming\\Microsoft\\Windows\\Start Menu\\Programs\\Startup".format(emu.get_user()['name']) # noqa
        elif csidl == 0x1e:
            path = "C:\\ProgramData\\Microsoft\\Windows\\Start Menu\\Programs\\Startup"
        elif csidl == 0x2a or csidl == 0x26:
            path = "C:\\Program Files"
        elif csidl == 0x2b or csidl == 0x2c:
            path = "C:\\Program Files\\Common Files"
        elif csidl == 0x24:
            path = "C:\\Windows"
        elif csidl == 0x25:
            path = "C:\\Windows\\System32"
        elif csidl == 0x14:
            path = "C:\\Windows\\Fonts"
        elif csidl == 0x23:
            path = "C:\\ProgramData"
        else:
            # Temp
            path = "C:\\Windows\\Temp"
        self.mem_write(pszPath, path.encode('utf-8'))
        return 0



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

from speakeasy.struct import EmuStruct
import ctypes as ct


USB_DEVICE_DESCRIPTOR_TYPE = 0x01
USB_CONFIGURATION_DESCRIPTOR_TYPE = 0x02
USB_STRING_DESCRIPTOR_TYPE = 0x03
USB_INTERFACE_DESCRIPTOR_TYPE = 0x04
USB_ENDPOINT_DESCRIPTOR_TYPE = 0x05

USB_ENDPOINT_TYPE_MASK = 0x03
USB_ENDPOINT_TYPE_CONTROL = 0x00
USB_ENDPOINT_TYPE_ISOCHRONOUS = 0x01
USB_ENDPOINT_TYPE_BULK = 0x02
USB_ENDPOINT_TYPE_INTERRUPT = 0x03

USB_DIR_IN = 0x80
USB_DIR_OUT = 0x00


class USB_DEVICE_DESCRIPTOR(EmuStruct):
    def __init__(self):
        super().__init__()
        self.bLength = ct.c_uint8
        self.bDescriptorType = ct.c_uint8
        self.bcdUSB = ct.c_uint16
        self.bDeviceClass = ct.c_uint8
        self.bDeviceSubClass = ct.c_uint8
        self.bDeviceProtocol = ct.c_uint8
        self.bMaxPacketSize0 = ct.c_uint8
        self.idVendor = ct.c_uint16
        self.idProduct = ct.c_uint16
        self.bcdDevice = ct.c_uint16
        self.iManufacturer = ct.c_uint8
        self.iProduct = ct.c_uint8
        self.iSerialNumber = ct.c_uint8
        self.bNumConfigurations = ct.c_uint8


class USB_CONFIGURATION_DESCRIPTOR(EmuStruct):
    def __init__(self):
        super().__init__()
        self.bLength = ct.c_uint8
        self.bDescriptorType = ct.c_uint8
        self.wTotalLength = ct.c_uint16
        self.bNumInterfaces = ct.c_uint8
        self.bConfigurationValue = ct.c_uint8
        self.iConfiguration = ct.c_uint8
        self.bmAttributes = ct.c_uint8
        self.MaxPower = ct.c_uint8


class USB_INTERFACE_DESCRIPTOR(EmuStruct):
    def __init__(self):
        super().__init__()
        self.bLength = ct.c_uint8
        self.bDescriptorType = ct.c_uint8
        self.bInterfaceNumber = ct.c_uint8
        self.bAlternateSetting = ct.c_uint8
        self.bNumEndpoints = ct.c_uint8
        self.bInterfaceClass = ct.c_uint8
        self.bInterfaceSubClass = ct.c_uint8
        self.bInterfaceProtocol = ct.c_uint8
        self.iInterface = ct.c_uint8


class USB_ENDPOINT_DESCRIPTOR(EmuStruct):
    def __init__(self):
        super().__init__()
        self.bLength = ct.c_uint8
        self.bDescriptorType = ct.c_uint8
        self.bEndpointAddress = ct.c_uint8
        self.bmAttributes = ct.c_uint8
        self.wMaxPacketSize = ct.c_uint16
        self.bInterval = ct.c_uint8


class USBD_VERSION_INFORMATION(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.USBDI_Version = ct.c_uint32
        self.Supported_USB_Version = ct.c_uint32



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

import base64

import speakeasy.winenv.defs.windows.windows as windefs

from .. import api


class Ncrypt(api.ApiHandler):
    """
    Implements exported functions from ncrypt.dll
    """
    name = 'ncrypt'
    apihook = api.ApiHandler.apihook
    impdata = api.ApiHandler.impdata

    def __init__(self, emu):

        super(Ncrypt, self).__init__(emu)

        self.funcs = {}
        self.data = {}

        super(Ncrypt, self).__get_hook_attrs__(self)

    @apihook('NCryptOpenStorageProvider', argc=3)
    def NCryptOpenStorageProvider(self, emu, argv, ctx={}):
        """
        SECURITY_STATUS NCryptOpenStorageProvider(
            NCRYPT_PROV_HANDLE *phProvider,
            LPCWSTR            pszProviderName,
            DWORD              dwFlags
        );
        """
        phProvider, pszProviderName, dwFlags = argv
        if pszProviderName:
            prov_str = self.read_wide_string(pszProviderName)
            argv[1] = prov_str

            cm = emu.get_crypt_manager()
            hnd = cm.crypt_open(pname=prov_str, flags=dwFlags)
            if hnd:
                self.mem_write(phProvider, hnd.to_bytes(emu.get_ptr_size(), 'little'))

        return windefs.ERROR_SUCCESS

    @apihook('NCryptImportKey', argc=8)
    def NCryptImportKey(self, emu, argv, ctx={}):
        """
        SECURITY_STATUS NCryptImportKey(
            NCRYPT_PROV_HANDLE hProvider,
            NCRYPT_KEY_HANDLE  hImportKey,
            LPCWSTR            pszBlobType,
            NCryptBufferDesc   *pParameterList,
            NCRYPT_KEY_HANDLE  *phKey,
            PBYTE              pbData,
            DWORD              cbData,
            DWORD              dwFlags
        );
        """
        hProvider, hImportKey, pszBlobType, pParameterList, phKey, pbData, cbData, dwFlags = argv
        blob_type = self.read_wide_string(pszBlobType)
        argv[2] = blob_type

        blob = self.mem_read(pbData, cbData)
        argv[5] = base64.b64encode(blob).decode('utf-8')

        cm = emu.get_crypt_manager()
        if hProvider and phKey:
            ctx = cm.crypt_get(hProvider)
            hnd = ctx.import_key(blob_type=blob_type,
                                 blob=blob,
                                 blob_len=cbData,
                                 hnd_import_key=hImportKey,
                                 param_list=pParameterList,
                                 flags=dwFlags)
            if hnd:
                self.mem_write(phKey, hnd.to_bytes(emu.get_ptr_size(), 'little'))

        return windefs.ERROR_SUCCESS

    @apihook('NCryptDeleteKey', argc=2)
    def NCryptDeleteKey(self, emu, argv, ctx={}):
        """
        SECURITY_STATUS NCryptDeleteKey(
            NCRYPT_KEY_HANDLE hKey,
            DWORD             dwFlags
        );
        """
        hKey, dwFlags = argv
        cm = emu.get_crypt_manager()
        for hnd, ctx in cm.ctx_handles.items():
            hnd_key = ctx.get_key(hKey)
            if hnd_key:
                ctx.delete_key(hKey)
                break

        return windefs.ERROR_SUCCESS

    @apihook('NCryptFreeObject', argc=1)
    def NCryptFreeObject(self, emu, argv, ctx={}):
        """
        SECURITY_STATUS NCryptFreeObject(
            NCRYPT_HANDLE hObject
        );
        """
        hObject = argv[0]
        cm = emu.get_crypt_manager()

        # hObject can be a handle to a provider or key
        for hnd, ctx in cm.ctx_handles.items():
            if hnd == hObject:
                cm.crypt_close(hObject)
                break
            elif ctx.get_key(hObject):
                ctx.delete_key(hObject)
                break

        return windefs.ERROR_SUCCESS



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

import speakeasy.winenv.arch as _arch

from .. import api


class Lz32(api.ApiHandler):

    name = 'lz32'
    apihook = api.ApiHandler.apihook
    impdata = api.ApiHandler.impdata

    def __init__(self, emu):

        super(Lz32, self).__init__(emu)
        super(Lz32, self).__get_hook_attrs__(self)

    @apihook('LZSeek', argc=3, conv=_arch.CALL_CONV_STDCALL)
    def LZSeek(self, emu, argv, ctx={}):
        """
        LONG LZSeek(
          INT  hFile,
          LONG lOffset,
          INT  iOrigin
        );
        """
        return -1



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

import random
import uuid

import speakeasy.winenv.defs.windows.windows as windefs

from .. import api

class RPCRT4(api.ApiHandler):
    """
    Implements exported functions from rpcrt4.dll
    """
    name = 'rpcrt4'
    apihook = api.ApiHandler.apihook
    impdata = api.ApiHandler.impdata

    def __init__(self, emu):
        super(RPCRT4, self).__init__(emu)

    @apihook('UuidCreate', argc=1)
    def UuidCreate(self, emu, argv, ctx={}):
        """
        RPC_STATUS UuidCreate(
          UUID *Uuid
        );
        """
        uuidp, = argv

        if not uuidp:
            return 1

        new_uuid = windefs.GUID()
        new_uuid.Data1 = random.randint(0, 0xffffffff)
        new_uuid.Data2 = random.randint(0, 0xffffffff) & 0xffff
        new_uuid.Data3 = random.randint(0, 0xffffffff) & 0xffff
        new_uuid.Data4 = random.randbytes(8)

        self.mem_write(uuidp, new_uuid.get_bytes())

        return 0
    
    @apihook('UuidToStringA', argc=2)
    def UuidToStringA(self, emu, argv, ctx={}):
        """
        RPC_STATUS UuidToStringA(
          const UUID *Uuid,
          RPC_CSTR   *StringUuid
        );
        """
        uuidp, stringp = argv

        if not uuidp or not stringp:
            return 1

        uuid_bytes = self.mem_read(uuidp, windefs.GUID().sizeof())
        uuid_obj = uuid.UUID(bytes=uuid_bytes)

        string = str(uuid_obj)

        self.mem_write(stringp, string.encode("utf-8"))

        return 0



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.


import collections

import speakeasy.winenv.arch as e_arch

# When disassembling, a minimum instruction size needs to be supplied
# This number is arbitrary and just needs to be large enough to cover
# the size of the current disasm target
DISASM_SIZE = 0x20


# default list of APIs to always allow despite triggering API hammering detection
_default_api_hammer_allowlist = [
    'kernel32.WriteProcessMemory',
    'kernel32.WriteFile',
    'kernel32.ReadFile',
]


def _lowercase_set(tt):
    return set([bb.lower() for bb in tt])


class ApiHammer():
    """
    Detect and attempt to mitigate API hammering as part of anti-sandbox or
    anti-emulation in malware samples
    """

    def __init__(self, emu):
        super(ApiHammer, self).__init__()
        self.emu = emu
        self.api_stats = collections.defaultdict(int)
        self.hammer_memregion = None
        self.hammer_offset = 0

        self.config = self.emu.config.get('api_hammering', {})
        self.api_threshold = self.config.get('threshold', 1000)
        self.enabled = self.config.get('enabled', False)
        self.allow_list = _lowercase_set(self.config.get('allow_list',
                                                         _default_api_hammer_allowlist))

    def is_allowed_api(self, apiname):
        '''
        Returns true if the given apiname is one we don't want to use api hammering
        mitigation for
        '''
        return apiname.lower() in self.allow_list

    def handle_import_func(self, imp_api, conv, argc):
        '''
        Identifies possible API hammering and attempts to patch in mitigations.
        '''
        if not self.enabled:
            # api hammering mitigation not enabled, so exit
            return
        if self.is_allowed_api(imp_api):
            # this is an api that we always want to allow, don't bother trying to
            # prevent api hammering
            return
        hammer_key = imp_api + '%x' % self.emu.get_ret_address()
        self.api_stats[hammer_key] += 1
        if self.api_stats[hammer_key] < self.api_threshold:
            return
        # TODO: better parameterize the checking & dispatch of the types of calls/jmps to imports
        # so we can more easily loop through them & clean up the the logic below
        # TODO: track patches in the hammer_memregion & reuse when possible
        if self.emu.get_arch() == e_arch.ARCH_X86:
            eip = self.emu.get_ret_address() - 6
            mnem, op, instr = self.emu.get_disasm(eip, DISASM_SIZE)
            self.emu.log_info('api hammering at: %s 0x%x %r %r %r' % (imp_api, self.emu.get_pc(),
                                                                      mnem, op, instr))
            if (mnem == 'call') and 'dword ptr' in instr:
                if conv == e_arch.CALL_CONV_CDECL:
                    # If cdecl, the emu engine will clean the stack
                    # just xor eax,eax & 4 bytes of nop
                    patch = b'\x31\xc0\x90\x90\x90\x90\x90'
                    self.emu.mem_write(eip, patch)
                    self.emu.log_info('API HAMMERING DETECTED - patching 1 cdecl at %x' % (eip, ))
                elif conv == e_arch.CALL_CONV_STDCALL:
                    # If stdcall, we need to clean the stack
                    # patch is xor eax, eax; add esp, <count>
                    patch = b'\x31\xc0\x83\xc4' + (4*argc).to_bytes(1, 'little') + b'\x90'
                    self.emu.mem_write(eip, patch)
                    self.emu.log_info('API HAMMERING DETECTED - patching 1 stdcall at %x' % (eip,))
            else:
                eip = self.emu.get_ret_address() - 2
                mnem, op, instr = self.emu.get_disasm(eip, DISASM_SIZE)
                self.emu.log_info('api hammering at: 0x%x %r %r %r' % (self.emu.get_pc(), mnem,
                                                                       op, instr))
                if (mnem == 'call') and op in e_arch.REG_LOOKUP.keys():
                    # not enough space to clean up stack inline, so write stack cleanup code to a
                    # hammerpatch region & change the register to point to this cleanup code
                    # instead the hope is that we're in a tight loop, so this will prevent exiting
                    # the emulator the majority of the time.
                    if self.hammer_memregion is None:
                        self.hammer_memregion = self.emu.mem_map(0x1024*4,
                                                                 tag='speakeasy.hammerpatch')
                    if conv == e_arch.CALL_CONV_CDECL:
                        # If cdecl, the emu engine will clean the stack
                        # just xor eax,eax; retn
                        patch = b'\x31\xc0\xc3'
                        self.emu.mem_write(eip, patch)
                        self.emu.log_info('API HAMMERING DETECTED - patching 2 cdecl at %x' % (eip,)) # noqa
                    elif conv == e_arch.CALL_CONV_STDCALL:
                        # patch is xor eax, eax; retn <count>
                        patch = b'\x31\xc0\xc2' + (4*argc).to_bytes(2, 'little') + b'\x90'
                        loc = self.hammer_memregion + self.hammer_offset
                        if (self.hammer_offset + len(patch)) < 0x1024*4:
                            self.emu.mem_write(loc, patch)
                            self.hammer_offset += len(patch)
                            # now change the the register
                            reg = e_arch.REG_LOOKUP[op]
                            self.emu.reg_write(reg, loc)
                            self.emu.log_info('API HAMMERING DETECTED - patching 2 stdcall at %x' % (eip,)) # noqa
                else:
                    self.emu.log_info('API HAMMERING DETECTED - unable to patch %x' % (eip, ))

        if self.emu.get_arch() == e_arch.ARCH_AMD64:
            # TODO
            pass



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

from urllib.parse import urlparse

import speakeasy.winenv.defs.windows.windows as windefs

from .. import api


class Urlmon(api.ApiHandler):

    name = 'urlmon'
    apihook = api.ApiHandler.apihook
    impdata = api.ApiHandler.impdata

    def __init__(self, emu):

        super(Urlmon, self).__init__(emu)
        self.funcs = {}
        self.data = {}
        super(Urlmon, self).__get_hook_attrs__(self)
        self.netman = emu.get_network_manager()
        self.names = {}

    @apihook('URLDownloadToFile', argc=5)
    def URLDownloadToFile(self, emu, argv, ctx={}):
        """
        HRESULT URLDownloadToFile(
                    LPUNKNOWN            pCaller,
                    LPCTSTR              szURL,
                    LPCTSTR              szFileName,
                    DWORD                dwReserved,
                    LPBINDSTATUSCALLBACK lpfnCB
        );
        """

        pCaller, szURL, szFileName, dwReserved, lpfnCB = argv
        rv = windefs.ERROR_SUCCESS

        cw = self.get_char_width(ctx)

        if szURL:
            url = self.read_mem_string(szURL, cw)
            argv[1] = url
            url = urlparse(url)
            if url.netloc:
                self.log_dns(url.netloc)

        if szFileName:
            name = self.read_mem_string(szFileName, cw)
            argv[2] = name
            self.log_file_access(name, 'create')
            self.log_file_access(name, 'write')

        return rv



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

from speakeasy.struct import EmuStruct, Ptr
import ctypes as ct

WH_CALLWNDPROC = 4
WH_CALLWNDPROCRET = 12
WH_CBT = 5
WH_DEBUG = 9
WH_FOREGROUNDIDLE = 11
WH_GETMESSAGE = 3
WH_JOURNALPLAYBACK = 1
WH_JOURNALRECORD = 0
WH_KEYBOARD = 2
WH_KEYBOARD_LL = 13
WH_MOUSE = 7
WH_MOUSE_LL = 14
WH_MSGFILTER = -1
WH_SHELL = 10
WH_SYSMSGFILTER = 6
WM_TIMER = 0x0113

WM_PAINT = 0x0F

WM_INITDIALOG = 0x0110


class MSG(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.hwnd = Ptr
        self.message = ct.c_uint32
        self.wParam = Ptr
        self.lParam = Ptr
        self.time = ct.c_uint32
        self.pt_x = Ptr
        self.pt_y = Ptr
        self.lPrivate = ct.c_uint32


class USEROBJECTFLAGS(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.fInherit = ct.c_uint32
        self.fReserved = ct.c_uint32
        self.dwFlags = ct.c_uint32


class WNDCLASSEX(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.cbSize = ct.c_uint32
        self.style = ct.c_uint32
        self.lpfnWndProc = Ptr
        self.cbClsExtra = ct.c_uint32
        self.cbWndExtra = ct.c_uint32
        self.hInstance = Ptr
        self.hIcon = Ptr
        self.hCursor = Ptr
        self.hbrBackground = Ptr
        self.lpszMenuName = Ptr
        self.lpszClassName = Ptr
        self.hIconSm = Ptr


def get_flag_defines(flags, prefix=''):
    defs = []
    for k, v in globals().items():
        if not isinstance(v, int):
            continue
        if v == flags:
            if prefix and k.startswith(prefix):
                defs.append(k)
    return defs


def get_windowhook_flags(flags):
    return get_flag_defines(flags, prefix='WH_')
